"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
import traceback
from typing import TYPE_CHECKING, Any, ClassVar

import ase.units
import numpy as np
import pandas as pd
from ase.filters import FrechetCellFilter
from tqdm import tqdm

from fairchem.core.components.calculate._calculate_runner import CalculateRunner
from fairchem.core.components.calculate.recipes.elastic import (
    calculate_elasticity,
)

if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator

    from fairchem.core.datasets.atoms_sequence import AtomsSequence


eVA3_to_GPa = 1 / ase.units.GPa


class ElasticityRunner(CalculateRunner):
    """Calculate elastic tensor for a set of structures."""

    result_glob_pattern: ClassVar[str] = "elasticity_*-*.json.gz"

    def __init__(
        self,
        calculator: Calculator,
        input_data: AtomsSequence,
    ):
        """
        Initialize the ElasticityRunner.

        Args:
            calculator: ASE calculator to use for energy and force calculations
            input_data: Dataset containing atomic structures to process
        """
        super().__init__(calculator=calculator, input_data=input_data)

    def calculate(self, job_num: int = 0, num_jobs: int = 1) -> list[dict[str, Any]]:
        """
        Calculate elastic properties for a batch of structures.

        Args:
            job_num (int, optional): Current job number in array job. Defaults to 0.
            num_jobs (int, optional): Total number of jobs in array. Defaults to 1.

        Returns:
            List of dictionaries containing elastic properties for each structure
        """
        all_results = []

        chunk_indices = np.array_split(range(len(self.input_data)), num_jobs)[job_num]
        for i in tqdm(chunk_indices, desc="Running elasticity calculations."):
            atoms = self.input_data[i]
            try:
                results = calculate_elasticity(
                    atoms,
                    calculator=self.calculator,
                    cell_filter_cls=FrechetCellFilter,
                    fix_symmetry=False,
                )
                results["sid"] = atoms.info.get("sid", i)
                # change results to GPa
                results["elastic_tensor"] = (
                    results["elastic_tensor"].voigt * eVA3_to_GPa
                )
                results["shear_modulus_vrh"] = (
                    results["shear_modulus_vrh"] * eVA3_to_GPa
                )
                results["bulk_modulus_vrh"] = results["bulk_modulus_vrh"] * eVA3_to_GPa
                results.update({"errors": "", "traceback": ""})
            except Exception as ex:
                results = {
                    "sid": atoms.info["sid"],
                    "elastic_tensor": np.nan,
                    "shear_modulus_vrh": np.nan,
                    "bulk_modulus_vrh": np.nan,
                    "errors": f"{ex!r}",
                    "traceback": traceback.format_exc(),
                }

            all_results.append(results)

        return all_results

    def write_results(
        self,
        results: list[dict[str, Any]],
        results_dir: str,
        job_num: int = 0,
        num_jobs: int = 1,
    ) -> None:
        """Write calculation results to a compressed JSON file.

        Args:
            results: List of dictionaries containing elastic properties
            results_dir: Directory path where results will be saved
            job_num: Index of the current job
            num_jobs: Total number of jobs
        """
        results_df = pd.DataFrame(results)
        results_df.to_json(
            os.path.join(results_dir, f"elasticity_{num_jobs}-{job_num}.json.gz")
        )

    def save_state(self, checkpoint_location: str, is_preemption: bool = False) -> bool:
        return True



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from fairchem.core.components.reducer import Reducer
    from fairchem.core.components.runner import Runner


class ManagedAttribute:
    """A descriptor helper to manage setting/access to an attribute of a class"""

    def __init__(self, enforced_type: type | None = None):
        self._enforced_type = enforced_type

    def __set_name__(self, owner: Runner | Reducer, name: str):
        self.public_name: str = name
        self.private_name: str = "_" + name

    def __get__(
        self, obj: Runner | Reducer, objtype: type[Runner | Reducer] | None = None
    ):
        return getattr(obj, self.private_name)

    def __set__(self, obj: Runner | Reducer, value: Any):
        if self._enforced_type is not None and not isinstance(
            value, self._enforced_type
        ):
            raise ValueError(
                f"{self.public_name} can only be set to an instance of {self._enforced_type} type!"
            )
        setattr(obj, self.private_name, value)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import ArrayLike
    from pymatgen.analysis.local_env import NearNeighbors
    from pymatgen.core import Structure


def construct_bond_matrix(
    structure: Structure,
    nn_finder: NearNeighbors,
    site_permutations: ArrayLike | None = None,
) -> np.ndarray:
    """
    Constructs a bond matrix for a given crystal structure.

    This function uses a near neigbor algorithm from pymatgen to determine bonds between atoms in the structure and
    creates an adjacency matrix where 1 indicates a bond between atoms and 0 indicates no bond.

    Args:
        structure: A pymatgen Structure object representing the crystal structure
        nn_finder: A NearNeighbors instance to determine near neighbor lists
        site_permutations: A numpy array containing the site permutations if the covalent matrix should be constructed
            with sites in an order different than the order in which they appear in the given structure

    Returns:
        np.ndarray: A square matrix where matrix[i,j] = 1 if atoms i and j share a covalent bond,
            and 0 otherwise
    """
    site_permutations = (
        site_permutations
        if site_permutations is not None
        else np.arange(len(structure), dtype=int)
    )

    nn_info = nn_finder.get_all_nn_info(structure)
    nn_matrix = np.zeros((len(nn_info), len(nn_info)), dtype=int)
    for i, ii in enumerate(site_permutations):
        for j in range(len(nn_info[i])):
            nn_matrix[ii, site_permutations[nn_info[i][j]["site_index"]]] = 1
    return nn_matrix



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging

from torch.utils.data import DataLoader, Dataset

from fairchem.core.common import distutils, gp_utils


def get_dataloader(
    dataset: Dataset, batch_sampler_fn: callable, collate_fn: callable, num_workers
) -> DataLoader:
    if gp_utils.initialized():
        num_replicas = gp_utils.get_dp_world_size()
        rank = gp_utils.get_dp_rank()
    else:
        num_replicas = distutils.get_world_size()
        rank = distutils.get_rank()

    logging.info(f"get_dataloader::Calling batch_sampler_fn={batch_sampler_fn}...")
    batch_sampler = batch_sampler_fn(
        dataset=dataset,
        num_replicas=num_replicas,
        rank=rank,
    )
    logging.info("get_dataloader::Calling Dataloader...")
    dataloader = DataLoader(
        dataset=dataset,
        collate_fn=collate_fn,
        num_workers=num_workers,
        pin_memory=True,
        batch_sampler=batch_sampler,
    )
    logging.info("get_dataloader::Done!")
    return dataloader



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

from functools import wraps
from typing import TYPE_CHECKING, Callable, ClassVar

import numpy as np
import torch

if TYPE_CHECKING:
    from collections.abc import Hashable

"""
An evaluation module for use with the OCP dataset and suite of tasks. It should
be possible to import this independently of the rest of the codebase, e.g:

```
from fairchem.core.modules import Evaluator

evaluator = Evaluator(task="is2re")
perf = evaluator.eval(prediction, target)
```

task: "s2ef", "is2rs", "is2re".

We specify a default set of metrics for each task, but should be easy to extend
to add more metrics. `evaluator.eval` takes as input two dictionaries, one for
predictions and another for targets to check against. It returns a dictionary
with the relevant metrics computed.
"""

NONE_SLICE = slice(None)


class Evaluator:
    task_metrics: ClassVar[dict[str, str]] = {
        "s2ef": {
            "energy": ["mae"],
            "forces": [
                "forcesx_mae",
                "forcesy_mae",
                "forcesz_mae",
                "mae",
                "cosine_similarity",
                "magnitude_error",
                "energy_forces_within_threshold",
            ],
        },
        "is2rs": {
            "positions": [
                "average_distance_within_threshold",
                "mae",
                "mse",
            ]
        },
        "is2re": {
            "energy": [
                "mae",
                "mse",
                "energy_within_threshold",
            ]
        },
    }

    task_primary_metric: ClassVar[dict[str, str | None]] = {
        "s2ef": "energy_forces_within_threshold",
        "is2rs": "average_distance_within_threshold",
        "is2re": "energy_mae",
        "ocp": None,
    }

    def __init__(
        self, task: str | None = None, eval_metrics: dict | None = None
    ) -> None:
        if eval_metrics is None:
            eval_metrics = {}
        self.task = task
        self.target_metrics = (
            eval_metrics if eval_metrics else self.task_metrics.get(task, {})
        )

    def eval(
        self,
        prediction: dict[str, torch.Tensor],
        target: dict[str, torch.Tensor],
        prev_metrics: dict | None = None,
    ):
        prev_metrics = prev_metrics or {}
        metrics = prev_metrics

        for target_property in self.target_metrics:
            for fn in self.target_metrics[target_property]:
                metric_name = (
                    f"{target_property}_{fn}"
                    if target_property not in fn and target_property != "misc"
                    else fn
                )
                res = eval(fn)(prediction, target, target_property)
                metrics = self.update(metric_name, res, metrics)

        return metrics

    def update(self, key, stat, metrics):
        if key not in metrics:
            metrics[key] = {
                "metric": None,
                "total": 0,
                "numel": 0,
            }

        if isinstance(stat, dict):
            # If dictionary, we expect it to have `metric`, `total`, `numel`.
            metrics[key]["total"] += stat["total"]
            metrics[key]["numel"] += stat["numel"]
            metrics[key]["metric"] = metrics[key]["total"] / metrics[key]["numel"]
        elif isinstance(stat, (float, int)):
            # If float or int, just add to the total and increment numel by 1.
            metrics[key]["total"] += stat
            metrics[key]["numel"] += 1
            metrics[key]["metric"] = metrics[key]["total"] / metrics[key]["numel"]
        elif torch.is_tensor(stat):
            raise NotImplementedError

        return metrics


def metrics_dict(metric_fun: Callable) -> Callable:
    """Wrap up the return of a metrics function"""

    @wraps(metric_fun)
    def wrapped_metrics(
        prediction: dict[str, torch.Tensor],
        target: dict[str, torch.Tensor],
        key: Hashable = None,
        **kwargs,
    ) -> dict[str, torch.Tensor]:
        error = metric_fun(prediction, target, key, **kwargs)
        return {
            "metric": torch.mean(error).item(),
            "total": torch.sum(error).item(),
            "numel": error.numel(),
        }

    return wrapped_metrics


@metrics_dict
def cosine_similarity(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = NONE_SLICE,
):
    # cast to float 32 to avoid 0/nan issues in fp16
    # https://github.com/pytorch/pytorch/issues/69512
    return torch.cosine_similarity(prediction[key].float(), target[key].float())


@metrics_dict
def mae(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = NONE_SLICE,
) -> torch.Tensor:
    return torch.abs(target[key] - prediction[key])


@metrics_dict
def mse(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = NONE_SLICE,
) -> torch.Tensor:
    return (target[key] - prediction[key]) ** 2


@metrics_dict
def per_atom_mae(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = NONE_SLICE,
) -> torch.Tensor:
    return torch.abs(target[key] - prediction[key]) / target["natoms"].unsqueeze(1)


@metrics_dict
def per_atom_mse(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = NONE_SLICE,
) -> torch.Tensor:
    return ((target[key] - prediction[key]) / target["natoms"].unsqueeze(1)) ** 2


@metrics_dict
def magnitude_error(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = NONE_SLICE,
    p: int = 2,
) -> torch.Tensor:
    assert prediction[key].shape[1] > 1
    return torch.abs(
        torch.norm(prediction[key], p=p, dim=-1) - torch.norm(target[key], p=p, dim=-1)
    )


def forcesx_mae(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = NONE_SLICE,
):
    return mae(prediction["forces"][:, 0], target["forces"][:, 0])


def forcesx_mse(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = NONE_SLICE,
):
    return mse(prediction["forces"][:, 0], target["forces"][:, 0])


def forcesy_mae(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = None,
):
    return mae(prediction["forces"][:, 1], target["forces"][:, 1])


def forcesy_mse(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = None,
):
    return mse(prediction["forces"][:, 1], target["forces"][:, 1])


def forcesz_mae(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = None,
):
    return mae(prediction["forces"][:, 2], target["forces"][:, 2])


def forcesz_mse(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = None,
):
    return mse(prediction["forces"][:, 2], target["forces"][:, 2])


def energy_forces_within_threshold(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = None,
) -> dict[str, float | int]:
    # Note that this natoms should be the count of free atoms we evaluate over.
    assert target["natoms"].sum() == prediction["forces"].size(0)
    assert target["natoms"].size(0) == prediction["energy"].size(0)

    # compute absolute error on per-atom forces and energy per system.
    # then count the no. of systems where max force error is < 0.03 and max
    # energy error is < 0.02.
    f_thresh = 0.03
    e_thresh = 0.02

    success = 0
    total = int(target["natoms"].size(0))

    error_forces = torch.abs(target["forces"] - prediction["forces"])
    error_energy = torch.abs(target["energy"] - prediction["energy"])

    start_idx = 0
    for i, n in enumerate(target["natoms"]):
        if (
            error_energy[i] < e_thresh
            and error_forces[start_idx : start_idx + n].max() < f_thresh
        ):
            success += 1
        start_idx += n

    return {
        "metric": success / total,
        "total": success,
        "numel": total,
    }


def energy_within_threshold(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = None,
) -> dict[str, float | int]:
    # compute absolute error on energy per system.
    # then count the no. of systems where max energy error is < 0.02.
    e_thresh = 0.02
    error_energy = torch.abs(target["energy"] - prediction["energy"])

    success = (error_energy < e_thresh).sum().item()
    total = target["energy"].size(0)

    return {
        "metric": success / total,
        "total": success,
        "numel": total,
    }


def average_distance_within_threshold(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = None,
) -> dict[str, float | int]:
    pred_pos = torch.split(prediction["positions"], prediction["natoms"].tolist())
    target_pos = torch.split(target["positions"], target["natoms"].tolist())

    mean_distance = []
    for idx, ml_pos in enumerate(pred_pos):
        mean_distance.append(
            np.mean(
                np.linalg.norm(
                    min_diff(
                        ml_pos.detach().cpu().numpy(),
                        target_pos[idx].detach().cpu().numpy(),
                        target["cell"][idx].detach().cpu().numpy(),
                        target["pbc"].tolist(),
                    ),
                    axis=1,
                )
            )
        )

    success = 0
    intv = np.arange(0.01, 0.5, 0.001)
    for i in intv:
        success += sum(np.array(mean_distance) < i)

    total = len(mean_distance) * len(intv)

    return {"metric": success / total, "total": success, "numel": total}


def min_diff(
    pred_pos: torch.Tensor,
    dft_pos: torch.Tensor,
    cell: torch.Tensor,
    pbc: torch.Tensor,
):
    pos_diff = pred_pos - dft_pos
    fractional = np.linalg.solve(cell.T, pos_diff.T).T

    for i, periodic in enumerate(pbc):
        # Yes, we need to do it twice
        if periodic:
            fractional[:, i] %= 1.0
            fractional[:, i] %= 1.0

    fractional[fractional > 0.5] -= 1

    return np.matmul(fractional, cell)


def rmse(
    prediction: dict[str, torch.Tensor],
    target: dict[str, torch.Tensor],
    key: Hashable = None,
) -> dict[str, float | int]:
    error = torch.sqrt(((target[key] - prediction[key]) ** 2).sum(dim=-1))
    return {
        "metric": torch.mean(error).item(),
        "total": torch.sum(error).item(),
        "numel": error.numel(),
    }



from __future__ import annotations

import logging
import os
import random
from typing import TYPE_CHECKING

import hydra
import numpy as np
import torch
from omegaconf import DictConfig, OmegaConf
from submitit import AutoExecutor
from submitit.core.utils import JobPaths, cloudpickle_dump
from submitit.helpers import Checkpointable, DelayedSubmission
from submitit.slurm.slurm import SlurmJobEnvironment

from fairchem.core.common import distutils
from fairchem.core.common.gp_utils import setup_graph_parallel_groups
from fairchem.core.common.logger import WandBSingletonLogger
from fairchem.core.common.utils import (
    setup_env_vars,
    setup_logging,
)
from fairchem.core.launchers.api import (
    DeviceType,
    JobConfig,
    RunType,
    SchedulerType,
    SlurmEnv,
)

if TYPE_CHECKING:
    from fairchem.core.components.reducer import Reducer
    from fairchem.core.components.runner import Runner


def _get_slurm_env() -> SlurmEnv:
    slurm_job_env = SlurmJobEnvironment()
    try:
        slurm_env = SlurmEnv(
            job_id=slurm_job_env.job_id,
            raw_job_id=slurm_job_env.raw_job_id,
            array_job_id=slurm_job_env.array_job_id,
            array_task_id=slurm_job_env.array_task_id,
            restart_count=os.environ.get("SLURM_RESTART_COUNT"),
        )
    except KeyError:
        # slurm environment variables are undefined, running locally
        slurm_env = SlurmEnv()

    return slurm_env


def map_job_config_to_dist_config(job_cfg: JobConfig) -> dict:
    scheduler_config = job_cfg.scheduler
    return {
        "world_size": scheduler_config.num_nodes * scheduler_config.ranks_per_node,
        "distributed_backend": (
            "gloo" if job_cfg.device_type == DeviceType.CPU else "nccl"
        ),
        "submit": scheduler_config.mode == SchedulerType.SLURM,
        "cpu": job_cfg.device_type == DeviceType.CPU,
        "init_method": scheduler_config.distributed_init_method,
        # for distributed shared file initialization
        "shared_file_dir": os.path.join(job_cfg.run_dir, job_cfg.timestamp_id),
        "array_job_num": job_cfg.metadata.array_job_num,
    }


def remove_runner_state_from_submission(log_folder: str, job_id: str) -> None:
    # (HACK) Decouple the job from the runner state by manually modifying it
    # this ensures the saved runner state is not re-submitted in the event of a node failure
    # ie: if the job was started at state t=T, a requeue during node failure would resubmit the job
    # starting at state t=T again without calling the checkpoint callback, losing all progress in between.
    job_path = JobPaths(folder=log_folder, job_id=job_id)
    if os.path.isfile(job_path.submitted_pickle):
        submission_obj = DelayedSubmission.load(job_path.submitted_pickle)
        submission_obj.args[0].job.runner_state_path = None
        cloudpickle_dump(submission_obj, job_path.submitted_pickle)


def runner_wrapper(config: DictConfig, run_type: RunType = RunType.RUN):
    # This is needed when using elastic_launch for local runs since it looks for
    # the __name__ attribute of the function, Submitit.__call__ does not have one
    SlurmSPMDProgram()(config, run_type)


def _set_seeds(seed: int) -> None:
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)


def _set_deterministic_mode() -> None:
    # this is required for full cuda deterministic mode
    logging.info("Setting deterministic mode!")
    os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"
    torch.backends.cudnn.deterministic = True
    torch.backends.cudnn.benchmark = False
    torch.use_deterministic_algorithms(True)


class SlurmSPMDProgram(Checkpointable):
    """
    Entrypoint for a SPMD program launched via submitit on slurm.
    This assumes all ranks run the identical copy of this code
    """

    def __init__(self) -> None:
        self.config = None
        self.runner = None
        self.reducer = None

    def __call__(
        self, dict_config: DictConfig, run_type: RunType = RunType.RUN
    ) -> None:
        self.config = dict_config
        self.run_type = run_type
        # modify the config metadata to add slurm info if they exist
        self.config.job.metadata.slurm_env = _get_slurm_env()

        setup_env_vars()
        setup_logging()

        dist_config = map_job_config_to_dist_config(self.config.job)
        logging.info("Setting up distributed backend...")
        distutils.setup(dist_config)
        distutils.synchronize()
        if (
            distutils.is_master()
            and self.config.job.scheduler.mode == SchedulerType.SLURM
        ):
            # this pickle file is shared across all processes so can only modify this on the main rank
            remove_runner_state_from_submission(
                dict_config.job.metadata.log_dir,
                self.config.job.metadata.slurm_env.job_id,
            )

        if self.config.job.graph_parallel_group_size is not None:
            logging.info("Setting up graph parallel...")
            setup_graph_parallel_groups(
                self.config.job.graph_parallel_group_size,
                dist_config["distributed_backend"],
            )

        self._init_logger()

        _set_seeds(self.config.job.seed)
        if self.config.job.deterministic:
            _set_deterministic_mode()

        if run_type == RunType.RUN:
            logging.info("Calling runner.run() ...")
            self.runner: Runner = hydra.utils.instantiate(self.config.runner)
            self.runner.job_config = self.config.job
            # must call resume state AFTER the runner has been initialized
            self.runner.load_state(self.config.job.runner_state_path)
            self.runner.run()
        elif run_type == RunType.REDUCE:
            logging.info("Calling reducer.reduce() ...")
            self.reducer: Reducer = hydra.utils.instantiate(self.config.reducer)
            self.reducer.job_config = self.config.job
            self.reducer.runner_config = self.config.runner
            # must call resume state AFTER the runner has been initialized
            self.reducer.load_state(self.config.job.runner_state_path)
            self.reducer.reduce()
        else:
            raise ValueError(f"run type {run_type} is not recognized!")

        distutils.cleanup()

    def _init_logger(self) -> None:
        if (
            self.config.job.logger
            and distutils.is_master()
            and not self.config.job.debug
            and self.config.job.metadata.array_job_num == 0
        ):
            # get a partial function from the config and instantiate wandb with it
            # currently code assumes that we only use the WandBSingletonLogger
            logger_initializer = hydra.utils.instantiate(self.config.job.logger)
            simple_config = OmegaConf.to_container(
                self.config, resolve=True, throw_on_missing=True
            )
            logger_initializer(
                config=simple_config,
                run_id=self.config.job.timestamp_id,
                run_name=self.config.job.run_name,
                log_dir=self.config.job.metadata.log_dir,
            )

    def checkpoint(self, *args, **kwargs) -> DelayedSubmission:
        logging.error("Submitit checkpointing callback is triggered")
        save_path = self.config.job.metadata.preemption_checkpoint_dir
        cfg_copy = self.config.copy()
        # only assign if the save was successful
        cfg_copy.job.runner_state_path = None

        if (
            self.run_type == RunType.RUN
            and self.runner.save_state(save_path, is_preemption=True)
        ) or (
            self.run_type == RunType.REDUCE
            and self.reducer.save_state(save_path, is_preemption=True)
        ):
            cfg_copy.job.runner_state_path = save_path

        if WandBSingletonLogger.initialized():
            WandBSingletonLogger.get_instance().mark_preempting()
        logging.info(
            f"Submitit checkpointing callback is completed, resuming with use the following state: {save_path}"
        )
        return DelayedSubmission(SlurmSPMDProgram(), cfg_copy)


def slurm_launch(cfg: DictConfig, log_dir: str) -> None:
    scheduler_cfg = cfg.job.scheduler
    executor = AutoExecutor(folder=log_dir, slurm_max_num_timeout=3)
    executor.update_parameters(
        name=cfg.job.run_name,
        mem_gb=scheduler_cfg.slurm.mem_gb,
        timeout_min=scheduler_cfg.slurm.timeout_hr * 60,
        slurm_partition=scheduler_cfg.slurm.partition,
        gpus_per_node=scheduler_cfg.ranks_per_node,
        cpus_per_task=scheduler_cfg.slurm.cpus_per_task,
        tasks_per_node=scheduler_cfg.ranks_per_node,
        nodes=scheduler_cfg.num_nodes,
        slurm_qos=scheduler_cfg.slurm.qos,
        slurm_account=scheduler_cfg.slurm.account,
        slurm_additional_parameters=scheduler_cfg.slurm.additional_parameters,
    )
    if scheduler_cfg.num_array_jobs == 1:
        job = executor.submit(SlurmSPMDProgram(), cfg)
        logging.info(
            f"Submitted job id: {cfg.job.timestamp_id}, slurm id: {job.job_id}, logs: {cfg.job.metadata.log_dir}"
        )
        jobs = [job]
    elif scheduler_cfg.num_array_jobs > 1:
        executor.update_parameters(
            slurm_array_parallelism=scheduler_cfg.num_array_jobs,
        )

        jobs = []
        with executor.batch():
            for job_number in range(scheduler_cfg.num_array_jobs):
                _cfg = cfg.copy()
                _cfg.job.metadata.array_job_num = job_number
                job = executor.submit(SlurmSPMDProgram(), _cfg)
                jobs.append(job)
        logging.info(f"Submitted {len(jobs)} jobs: {jobs[0].job_id.split('_')[0]}")

    if "reducer" in cfg:
        job_id = jobs[0].job_id.split("_")[0]
        executor.update_parameters(
            name=f"{cfg.job.run_name}_reduce",
            # set a single node, or do we want the same config as the Runner or a separate JobConfig
            nodes=1,
            slurm_dependency=f"afterok:{job_id}",
            slurm_additional_parameters={
                "kill-on-invalid-dep": "yes"
            },  # kill the reducer if run fails
        )
        executor.submit(SlurmSPMDProgram(), cfg, RunType.REDUCE)


def local_launch(cfg: DictConfig, log_dir: str):
    """
    Launch locally with torch elastic (for >1 workers) or just single process
    """
    scheduler_cfg = cfg.job.scheduler
    if scheduler_cfg.ranks_per_node > 1:
        from torch.distributed.launcher.api import LaunchConfig, elastic_launch

        launch_config = LaunchConfig(
            min_nodes=1,
            max_nodes=1,
            nproc_per_node=scheduler_cfg.ranks_per_node,
            rdzv_backend="c10d",
            max_restarts=0,
        )
        elastic_launch(launch_config, runner_wrapper)(cfg)
        if "reducer" in cfg:
            elastic_launch(launch_config, runner_wrapper)(cfg, RunType.REDUCE)
    else:
        logging.info("Running in local mode without elastic launch")
        distutils.setup_env_local()
        runner_wrapper(cfg)
        if "reducer" in cfg:
            runner_wrapper(cfg, RunType.REDUCE)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from .scale_factor import ScaleFactor

if TYPE_CHECKING:
    import torch.nn as nn


def ensure_fitted(module: nn.Module, warn: bool = False) -> None:
    for name, child in module.named_modules():
        if not isinstance(child, ScaleFactor) or child.fitted:
            continue
        msg = (
            f"Scale factor {f'{child.name} ({name})' if child.name is not None else name} is not fitted. "
            "Please make sure that you either (1) load a checkpoint with fitted scale factors, "
            "(2) explicitly load scale factors using the `model.scale_file` attribute, or "
            "(3) fit the scale factors using the `fit.py` script."
        )
        if warn:
            logging.warning(msg)
        else:
            raise ValueError(msg)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

Helper scripts to run phonon calculations

- Compute phonon frequencies at commensurate points
- Compute thermal properties with Fourier interpolation
- Optionally compute and plot band-structures and DOS

Needs phonopy installed
"""

from __future__ import annotations

from typing import TYPE_CHECKING

import ase.units
import numpy as np
from ase import Atoms
from ase.build.supercells import make_supercell
from ase.constraints import FixSymmetry
from ase.filters import FrechetCellFilter
from ase.optimize import FIRE
from monty.dev import requires

try:
    from pymatgen.core import Structure
    from pymatgen.io.phonopy import get_phonopy_structure, get_pmg_structure

    pmg_installed = True
except ImportError:
    pmg_installed = False

try:
    from phonopy import Phonopy
    from phonopy.harmonic.dynmat_to_fc import get_commensurate_points

    phonopy_installed = True
except ImportError:
    phonopy_installed = False


if TYPE_CHECKING:
    from ase.calculators.calculator import Calculator
    from numpy.typing import ArrayLike, NDArray
    from phonopy.structure.atoms import PhonopyAtoms


THz_to_K = ase.units._hplanck * 1e12 / ase.units._k


@requires(phonopy_installed, message="Requires `phonopy` to be installed")
@requires(pmg_installed, message="Requires `pymatgen` to be installed")
def run_mdr_phonon_benchmark(
    mdr_phonon: Phonopy,
    calculator: Calculator,
    displacement: float = 0.01,
    run_relax: bool = True,
    fix_symm_relax: bool = False,
    symprec: int = 1e-4,
    symmetrize_fc: bool = False,
) -> dict:
    """Run a phonon calculation for a single datapoint of the MDR PBE dataset

    Properties computed for benchmark:
        - maximum frequency from phonon frequencies computed at supercell commensurate points
        - vibrational free energy, entropy and heat capacity computed with a [20, 20, 20] mesh

    Args:
        mdr_phonon: the baseline MDR Phonopy object
        calculator: an Ase Calculator
        displacement: displacement step to compute forces (A)
        run_relax: run a structural relaxation
        fix_symm_relax: wether to fix symmetry in relaxation
        symprec: symmetry precision used by phonopy
        symmetrize_fc: symmetrize force constants

    Returns:
        dict: dictionary of computed properties
    """

    if run_relax:
        # relax the primitive cell instead of the unitcell for efficiency
        primcell = get_pmg_structure(mdr_phonon.primitive).to_ase_atoms()

        if fix_symm_relax:
            primcell.set_constraint(FixSymmetry(primcell))

        primcell.calc = calculator
        opt = FIRE(FrechetCellFilter(primcell), logfile=None)
        opt.run(fmax=0.005, steps=500)
        natoms = len(primcell.positions)
        final_energy_per_atom = primcell.get_potential_energy() / natoms
        final_volume_per_atom = primcell.get_volume() / natoms
        if mdr_phonon.primitive_matrix is not None:
            P = np.asarray(np.linalg.inv(mdr_phonon.primitive_matrix.T), dtype=np.intc)
            unitcell = make_supercell(primcell, P)
        else:  # assume prim is the same as unit
            # can always check for good measure
            # assert np.allclose(mdr_phonon.unitcell, mdr_phonon.primitive.cell)
            unitcell = primcell
    else:
        unitcell = mdr_phonon.unitcell
        final_energy_per_atom = np.nan
        final_volume_per_atom = np.nan

    phonon = get_phonopy_object(
        unitcell,
        displacement=displacement,
        supercell_matrix=mdr_phonon.supercell_matrix,
        primitive_matrix=mdr_phonon.primitive_matrix,
        symprec=symprec,
    )
    produce_force_constants(phonon, calculator, symmetrize=symmetrize_fc)

    results = {
        "frequencies": calculate_phonon_frequencies(phonon) * THz_to_K,
        "energy_per_atom": final_energy_per_atom,
        "volume_per_atom": final_volume_per_atom,
        **calculate_thermal_properties(phonon, t_step=75, t_max=600, t_min=0),
    }

    return results


def get_phonopy_object(
    atoms: PhonopyAtoms | Atoms | Structure,
    displacement: float = 0.01,
    supercell_matrix: ArrayLike = ((2, 0, 0), (0, 2, 0), (0, 0, 2)),
    primitive_matrix: ArrayLike | None = None,
    symprec: int = 1e-5,
    **phonopy_kwargs,
) -> Phonopy:
    """Create a Phonopy api object from ase Atoms.

    Args:
        atoms: Phonopy atoms, ASE atoms object or a pmg Structure
        displacement: displacement step to compute forces (A)
        supercell_matrix: transformation matrix to super cell from unit cell.
        primitive_matrix: transformation matrix to primitive cell from unit cell.
        symprec: symmetry precision
        phonopy_kwargs: additional keyword arguments to initialize Phonopy API object
    Returns:
        Phonopy: api object
    """
    if isinstance(atoms, Atoms):
        atoms = Structure.from_ase_atoms(atoms)

    if isinstance(atoms, Structure):
        atoms = get_phonopy_structure(atoms)

    supercell_matrix = np.ascontiguousarray(supercell_matrix, dtype=int)
    phonon = Phonopy(
        atoms,
        supercell_matrix,
        primitive_matrix=primitive_matrix,
        symprec=symprec,
        **phonopy_kwargs,
    )
    phonon.generate_displacements(distance=displacement)
    return phonon


def produce_force_constants(
    phonon: Phonopy, calculator: Calculator, symmetrize: bool = False
) -> None:
    """Run force calculations and produce force constants with Phonopy

    Args:
        phonon: a Phonopy API object
        calculator: an ASE Calculator
        symmetrize: symmetrize force constants
    """

    phonon.forces = [
        calculator.get_forces(get_pmg_structure(supercell).to_ase_atoms())
        for supercell in phonon.supercells_with_displacements
    ]
    phonon.produce_force_constants()

    if symmetrize:
        phonon.symmetrize_force_constants()
        phonon.symmetrize_force_constants_by_space_group()


def calculate_phonon_frequencies(
    phonon: Phonopy, qpoints: ArrayLike | None = None
) -> NDArray:
    """Calculate phonon frequencies at a given set of qpoints.

    Args:
        phonon: a Phonopy api object with displacements generated
        qpoints: ndarray of qpoints to calculate phonon frequencies at. If none are given, the supercell commensurate
            points will be used

    Returns:
        NDArray: ndarray of phonon frequencies in THz, (qpoints, frequencies)
    """
    if qpoints is None:
        qpoints = get_commensurate_points(phonon.supercell_matrix)

    frequencies = np.stack([phonon.get_frequencies(q) for q in qpoints])

    return frequencies


def calculate_thermal_properties(
    phonon: Phonopy, t_min, t_max, t_step, mesh: ArrayLike = (20, 20, 20)
) -> dict[str, float]:
    """Calculate thermal properties from initialized phonopy object

    Thermal properties include: vibrational free energy, entropy and heat capacity

    Args:
        phonon: a Phonopy api object with displacements generated
        t_min: minimum temperature
        t_max: max temperature
        t_step: temperature step between min and max
        mesh: qpoint mesh to compute properties using Fourier interpolation

    Returns:
        dict: dictionary of computed properties
    """

    phonon.run_mesh(mesh)
    phonon.run_thermal_properties(t_min=t_min, t_max=t_max, t_step=t_step)
    return phonon.get_thermal_properties_dict()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
from typing import TYPE_CHECKING

import torch
from torch.profiler import ProfilerActivity, profile
from torchtnt.framework.callback import Callback

from fairchem.core.common import distutils
from fairchem.core.common.logger import WandBSingletonLogger

if TYPE_CHECKING:
    from omegaconf import DictConfig
    from torchtnt.framework import State, TTrainUnit

    from fairchem.core.common.logger import Logger


def get_default_profiler_handler(
    run_id: str, output_dir: str, logger: Logger, all_ranks: bool = False
):
    """Get a standard callback handle for the pytorch profiler"""

    def trace_handler(p):
        if all_ranks or distutils.is_master():
            trace_name = f"{run_id}_rank_{distutils.get_rank()}.pt.trace.json"
            output_path = os.path.join(output_dir, trace_name)
            logging.info(f"Saving trace in {output_path}")
            p.export_chrome_trace(output_path)
            if logger:
                logger.log_artifact(
                    name=trace_name, type="profile", file_location=output_path
                )

    return trace_handler


def get_profile_schedule(wait: int = 5, warmup: int = 5, active: int = 2):
    """Get a profile schedule and total number of steps to run
    check pytorch docs on the meaning of these paramters:
    https://pytorch.org/docs/stable/profiler.html#torch.profiler.schedule
    Example usage:
    ```
    trace_handler = get_default_profiler_handler(run_id = self.config["cmd"]["timestamp_id"],
                                                    output_dir = self.config["cmd"]["results_dir"],
                                                    logger = self.logger)
    profile_schedule, total_profile_steps = get_profile_schedule()

    with profile(
        activities=[ProfilerActivity.CPU, ProfilerActivity.CUDA],
        schedule=profile_schedule,
        on_trace_ready=trace_handler
    ) as p:
        for i in steps:
            <code block to profile>
            if i < total_profile_steps:
                p.step()
    """
    total_profile_steps = wait + warmup + active
    profile_schedule = torch.profiler.schedule(wait=wait, warmup=warmup, active=active)

    return profile_schedule, total_profile_steps


class ProfilerCallback(Callback):
    def __init__(
        self,
        job_config: DictConfig,
        wait_steps: int = 5,
        warmup_steps: int = 5,
        active_steps: int = 2,
        all_ranks: bool = False,
        activities: tuple = (ProfilerActivity.CPU, ProfilerActivity.CUDA),
    ) -> None:
        profile_dir = os.path.join(job_config.metadata.log_dir, "profiles")
        os.makedirs(profile_dir, exist_ok=True)
        logger = (
            WandBSingletonLogger.get_instance()
            if WandBSingletonLogger.initialized()
            else None
        )
        handler = get_default_profiler_handler(
            run_id=job_config.run_name,
            output_dir=profile_dir,
            logger=logger,
            all_ranks=all_ranks,
        )
        schedule, self.total_steps = get_profile_schedule(
            wait_steps, warmup_steps, active_steps
        )
        self.profiler = profile(
            activities=activities, schedule=schedule, on_trace_ready=handler
        )

    def on_train_start(self, state: State, unit: TTrainUnit) -> None:
        self.profiler.start()

    def on_train_step_end(self, state: State, unit: TTrainUnit) -> None:
        step = unit.train_progress.num_steps_completed
        if step <= self.total_steps:
            self.profiler.step()
        else:
            self.profiler.stop()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
import os
import subprocess
from datetime import timedelta
from typing import Any, TypeVar

import torch
import torch.distributed as dist
from torch.distributed.elastic.utils.distributed import get_free_port
from torchtnt.utils.distributed import get_file_init_method, get_tcp_init_method

from fairchem.core.common.typing import none_throws

T = TypeVar("T")
DISTRIBUTED_PORT = 13356
CURRENT_DEVICE_TYPE_STR = "CURRRENT_DEVICE_TYPE"


def os_environ_get_or_throw(x: str) -> str:
    if x not in os.environ:
        raise RuntimeError(f"Could not find {x} in ENV variables")
    return none_throws(os.environ.get(x))


def get_init_method(
    init_method,
    world_size: int | None,
    rank: int | None = None,
    node_list: str | None = None,
    filename: str | None = None,
):
    """
    Get the initialization method for a distributed job based on the specified method type.

    Args:
        init_method: The initialization method type, either "tcp" or "file".
        world_size: The total number of processes in the distributed job.
        rank: The rank of the current process (optional).
        node_list: The list of nodes for SLURM-based distributed job (optional, used with "tcp").
        filename: The shared file path for file-based initialization (optional, used with "file").

    Returns:
        The initialization method string to be used by PyTorch's distributed module.

    Raises:
        ValueError: If an invalid init_method is provided.
    """
    if init_method == "tcp":
        hostnames = subprocess.check_output(
            ["scontrol", "show", "hostnames", node_list]
        )
        return get_tcp_init_method(
            world_size=world_size,
            hostname=hostnames.split()[0].decode("utf-8"),
            rank=rank,
            port=get_free_port() if world_size == 1 else DISTRIBUTED_PORT,
        )
    elif init_method == "file":
        return get_file_init_method(world_size=world_size, rank=rank, filename=filename)
    else:
        raise ValueError(f"Invalid init_method: {init_method}")


def setup(config) -> None:
    timeout = timedelta(minutes=config.get("timeout", 30))
    if config["submit"]:
        node_list = os.environ.get("SLURM_STEP_NODELIST")
        if node_list is None:
            node_list = os.environ.get("SLURM_JOB_NODELIST")
        if node_list is not None:
            try:
                nnodes = int(os_environ_get_or_throw("SLURM_NNODES"))
                ntasks_per_node = os.environ.get("SLURM_NTASKS_PER_NODE")
                if ntasks_per_node is not None:
                    ntasks_per_node = int(ntasks_per_node)
                else:
                    ntasks = int(os_environ_get_or_throw("SLURM_NTASKS"))
                    nnodes = int(os_environ_get_or_throw("SLURM_NNODES"))
                    assert ntasks % nnodes == 0
                    ntasks_per_node = int(ntasks / nnodes)
                if ntasks_per_node == 1:
                    assert config["world_size"] % nnodes == 0
                    gpus_per_node = config["world_size"] // nnodes
                    node_id = int(os_environ_get_or_throw("SLURM_NODEID"))
                    rank = node_id * gpus_per_node
                    local_rank = 0
                else:
                    assert ntasks_per_node == config["world_size"] // nnodes
                    rank = int(os_environ_get_or_throw("SLURM_PROCID"))
                    local_rank = int(os_environ_get_or_throw("SLURM_LOCALID"))

                init_method = get_init_method(
                    config["init_method"],
                    world_size=config["world_size"],
                    rank=rank,
                    node_list=node_list,
                    filename=os.path.join(
                        config["shared_file_dir"],
                        f".distributed-shared-file-{config['array_job_num']}",
                    ),
                )
                logging.info(f"Torch distributed initialized with: {init_method}")

                # ensures GPU0 does not have extra context/higher peak memory
                logging.info(
                    f"local rank: {local_rank}, visible devices: {os.environ.get('CUDA_VISIBLE_DEVICES', 'None')}"
                )

                assign_device_for_local_rank(config["cpu"], local_rank)

                dist.init_process_group(
                    backend="nccl",
                    init_method=init_method,
                    timeout=timeout,
                )
            except subprocess.CalledProcessError as e:  # scontrol failed
                raise e
            except FileNotFoundError:  # Slurm is not installed
                pass
    else:  # local mode
        if "MASTER_ADDR" not in os.environ:
            assert (
                config["world_size"] == 1
            ), "Can only setup master address and port at this point for a single rank, otherwise we assume the processes and the comm addr/port have already been setup"
            setup_env_local()
        local_rank = int(os.environ["LOCAL_RANK"])
        assign_device_for_local_rank(config["cpu"], local_rank)

        dist.init_process_group(
            backend=config["distributed_backend"],
            rank=int(os.environ["RANK"]),
            world_size=config["world_size"],
            timeout=timeout,
        )


def cleanup() -> None:
    if dist.is_initialized():
        dist.destroy_process_group()
    if CURRENT_DEVICE_TYPE_STR in os.environ:
        os.environ.pop(CURRENT_DEVICE_TYPE_STR)


def initialized() -> bool:
    return dist.is_available() and dist.is_initialized()


def get_rank() -> int:
    return dist.get_rank() if initialized() else 0


def get_world_size() -> int:
    return dist.get_world_size() if initialized() else 1


def is_master() -> bool:
    return get_rank() == 0


def synchronize() -> None:
    if get_world_size() == 1:
        return
    dist.barrier()


def broadcast(
    tensor: torch.Tensor, src, group=dist.group.WORLD, async_op: bool = False
) -> None:
    if get_world_size() == 1:
        return
    dist.broadcast(tensor, src, group, async_op)


def broadcast_object_list(
    object_list: list[Any], src: int, group=dist.group.WORLD, device: str | None = None
) -> None:
    if get_world_size() == 1:
        return
    dist.broadcast_object_list(object_list, src, group, device)


def all_reduce(
    data, group=dist.group.WORLD, average: bool = False, device=None
) -> torch.Tensor:
    if get_world_size() == 1:
        return data
    tensor = data
    if not isinstance(data, torch.Tensor):
        tensor = torch.tensor(data)
    if device is not None:
        tensor = tensor.to(device)
    dist.all_reduce(tensor, group=group)
    if average:
        tensor /= get_world_size()
    if not isinstance(data, torch.Tensor):
        result = tensor.cpu().numpy() if tensor.numel() > 1 else tensor.item()
    else:
        result = tensor
    return result


def all_gather(data, group=dist.group.WORLD, device=None) -> list[torch.Tensor]:
    if get_world_size() == 1:
        return data
    tensor = data
    if not isinstance(data, torch.Tensor):
        tensor = torch.tensor(data)
    if device is not None:
        tensor = tensor.to(device)
    tensor_list = [tensor.new_zeros(tensor.shape) for _ in range(get_world_size())]
    dist.all_gather(tensor_list, tensor, group=group)
    if not isinstance(data, torch.Tensor):
        result = [tensor.cpu().numpy() for tensor in tensor_list]
    else:
        result = tensor_list
    return result


def gather_objects(data: T, group: dist.ProcessGroup = dist.group.WORLD) -> list[T]:
    """Gather a list of pickleable objects into rank 0"""
    if get_world_size() == 1:
        return [data]

    output = [None for _ in range(get_world_size())] if is_master() else None
    dist.gather_object(data, output, group=group, dst=0)
    return output


def assign_device_for_local_rank(cpu: bool, local_rank: int) -> None:
    if cpu:
        os.environ[CURRENT_DEVICE_TYPE_STR] = "cpu"
    else:
        assert torch.cuda.is_available(), "cannot set cpu=false and no cuda available!"
        os.environ[CURRENT_DEVICE_TYPE_STR] = "cuda"
        torch.cuda.set_device(local_rank)


def get_device_for_local_rank() -> str:
    if os.environ.get(CURRENT_DEVICE_TYPE_STR) is None:
        os.environ[CURRENT_DEVICE_TYPE_STR] = (
            f"cuda:{torch.cuda.current_device()}"
            if torch.cuda.is_available()
            else "cpu"
        )
        logging.warning(
            f"WARNING: assign_device_for_local_rank was never called, automatically defaulting to using {os.environ[CURRENT_DEVICE_TYPE_STR]}"
        )
        return os.environ[CURRENT_DEVICE_TYPE_STR]

    if "cuda" in os.environ[CURRENT_DEVICE_TYPE_STR]:
        assert torch.cuda.is_available(), "cannot set cpu=false and no cuda available!"
        return f"cuda:{torch.cuda.current_device()}"
    elif os.environ[CURRENT_DEVICE_TYPE_STR] == "cpu":
        return "cpu"
    else:
        raise ValueError(f"unsupported device type: {CURRENT_DEVICE_TYPE_STR}")


def setup_env_local():
    os.environ["MASTER_ADDR"] = "localhost"
    os.environ["LOCAL_RANK"] = "0"
    os.environ["RANK"] = "0"
    os.environ["MASTER_PORT"] = str(get_free_port())


def setup_env_local_multi_gpu(rank: int, port: int, address: str = "localhost"):
    os.environ["MASTER_ADDR"] = address
    os.environ["LOCAL_RANK"] = str(rank)
    os.environ["RANK"] = str(rank)
    os.environ["MASTER_PORT"] = str(port)



from __future__ import annotations

import numpy as np
from ase.build import bulk


def get_fcc_carbon_xtal(
    num_atoms: int,
    lattice_constant: float = 3.8,
):
    # lattice_constant = 3.8, fcc generates a supercell with ~50 edges/atom, used for benchmarking
    atoms = bulk("C", "fcc", a=lattice_constant)
    n_cells = int(np.ceil(np.cbrt(num_atoms)))
    atoms = atoms.repeat((n_cells, n_cells, n_cells))
    indices = np.random.choice(len(atoms), num_atoms, replace=False)
    sampled_atoms = atoms[indices]
    return sampled_atoms



from __future__ import annotations

from functools import partial
from typing import TYPE_CHECKING

import torch
import torch.nn as nn

from fairchem.core.common.registry import registry
from fairchem.core.common.utils import conditional_grad
from fairchem.core.models.base import BackboneInterface, HeadInterface
from fairchem.core.models.escaip.configs import EScAIPConfigs, init_configs
from fairchem.core.models.escaip.modules.graph_attention_block import (
    EfficientGraphAttentionBlock,
)
from fairchem.core.models.escaip.modules.input_block import InputBlock
from fairchem.core.models.escaip.modules.output_block import (
    OutputLayer,
    OutputProjection,
)
from fairchem.core.models.escaip.modules.readout_block import ReadoutBlock
from fairchem.core.models.escaip.utils.data_preprocess import (
    data_preprocess_radius_graph,
)
from fairchem.core.models.escaip.utils.graph_utils import (
    compilable_scatter,
    get_displacement_and_cell,
    unpad_results,
)
from fairchem.core.models.escaip.utils.nn_utils import (
    get_normalization_layer,
    init_linear_weights,
    no_weight_decay,
)

if TYPE_CHECKING:
    from fairchem.core.datasets.atomic_data import AtomicData
    from fairchem.core.models.escaip.custom_types import GraphAttentionData


@registry.register_model("EScAIP_backbone")
class EScAIPBackbone(nn.Module, BackboneInterface):
    """
    Efficiently Scaled Attention Interactomic Potential (EScAIP) backbone model.
    """

    def __init__(
        self,
        **kwargs,
    ):
        super().__init__()

        # load configs
        cfg = init_configs(EScAIPConfigs, kwargs)
        self.global_cfg = cfg.global_cfg
        self.molecular_graph_cfg = cfg.molecular_graph_cfg
        self.gnn_cfg = cfg.gnn_cfg
        self.reg_cfg = cfg.reg_cfg

        # for trainer
        self.regress_forces = cfg.global_cfg.regress_forces
        self.direct_forces = cfg.global_cfg.direct_forces
        self.regress_stress = cfg.global_cfg.regress_stress
        self.dataset_list = cfg.global_cfg.dataset_list
        self.max_num_elements = cfg.molecular_graph_cfg.max_num_elements
        self.max_neighbors = cfg.molecular_graph_cfg.knn_k
        self.cutoff = cfg.molecular_graph_cfg.max_radius

        # data preprocess
        self.data_preprocess = partial(
            data_preprocess_radius_graph,
            global_cfg=self.global_cfg,
            gnn_cfg=self.gnn_cfg,
            molecular_graph_cfg=self.molecular_graph_cfg,
        )

        ## Model Components

        # Input Block
        self.input_block = InputBlock(
            global_cfg=self.global_cfg,
            molecular_graph_cfg=self.molecular_graph_cfg,
            gnn_cfg=self.gnn_cfg,
            reg_cfg=self.reg_cfg,
        )

        # Transformer Blocks
        self.transformer_blocks = nn.ModuleList(
            [
                EfficientGraphAttentionBlock(
                    global_cfg=self.global_cfg,
                    molecular_graph_cfg=self.molecular_graph_cfg,
                    gnn_cfg=self.gnn_cfg,
                    reg_cfg=self.reg_cfg,
                    is_last=(idx == self.global_cfg.num_layers - 1),
                )
                for idx in range(self.global_cfg.num_layers)
            ]
        )

        # Readout Layer
        self.readout_layers = nn.ModuleList(
            [
                ReadoutBlock(
                    global_cfg=self.global_cfg,
                    gnn_cfg=self.gnn_cfg,
                    reg_cfg=self.reg_cfg,
                )
                for _ in range(self.global_cfg.num_layers + 1)
            ]
        )

        # Output Projection
        self.output_projection = OutputProjection(
            global_cfg=self.global_cfg,
            gnn_cfg=self.gnn_cfg,
            reg_cfg=self.reg_cfg,
        )

        # init weights
        self.init_weights()

        # enable torch.set_float32_matmul_precision('high')
        torch.set_float32_matmul_precision("high")

        # log recompiles
        torch._logging.set_logs(recompiles=True)  # type: ignore

    def compiled_forward(self, data: GraphAttentionData):
        # input block
        node_features, edge_features = self.input_block(data)

        # input readout
        readouts = self.readout_layers[0](data, node_features, edge_features)
        global_readouts = [readouts[0]]
        node_readouts = [readouts[1]]
        edge_readouts = [readouts[2]]

        # transformer blocks
        for idx in range(self.global_cfg.num_layers):
            node_features, edge_features = self.transformer_blocks[idx](
                data, node_features, edge_features
            )
            readouts = self.readout_layers[idx + 1](data, node_features, edge_features)
            readouts = self.readout_layers[idx + 1](data, node_features, edge_features)
            global_readouts.append(readouts[0])
            node_readouts.append(readouts[1])
            edge_readouts.append(readouts[2])

        global_features, node_features, edge_features = self.output_projection(
            data=data,
            global_readouts=torch.cat(global_readouts, dim=-1),
            node_readouts=torch.cat(node_readouts, dim=-1),
            edge_readouts=torch.cat(edge_readouts, dim=-1),
        )

        return {
            "data": data,
            "global_features": global_features.to(torch.float32)
            if global_features is not None
            else None,
            "node_features": node_features.to(torch.float32),
            "edge_features": edge_features.to(torch.float32)
            if edge_features is not None
            else None,
        }

    @conditional_grad(torch.enable_grad())
    def forward(self, data: AtomicData):
        # TODO: remove this when FairChem fixes the bug
        data["atomic_numbers"] = data["atomic_numbers"].long()  # type: ignore
        data["atomic_numbers_full"] = data["atomic_numbers"]  # type: ignore
        data["batch_full"] = data["batch"]  # type: ignore

        # gradient force and stress
        displacement, orig_cell = get_displacement_and_cell(
            data, self.regress_stress, self.regress_forces, self.direct_forces
        )

        # preprocess data
        x = self.data_preprocess(data)

        # compile forward function
        self.forward_fn = (
            torch.compile(self.compiled_forward)
            if self.global_cfg.use_compile
            else self.compiled_forward
        )

        results = self.forward_fn(x)
        results["displacement"] = displacement
        results["orig_cell"] = orig_cell
        return results

    @torch.jit.ignore(drop=False)
    def no_weight_decay(self):
        return no_weight_decay(self)

    def init_weights(self):
        for _, module in self.named_modules():
            if isinstance(module, nn.Linear) and module.bias is not None:
                module.bias.data.zero_()


class EScAIPHeadBase(nn.Module, HeadInterface):
    def __init__(self, backbone: EScAIPBackbone):  # type: ignore
        super().__init__()
        self.global_cfg = backbone.global_cfg
        self.molecular_graph_cfg = backbone.molecular_graph_cfg
        self.gnn_cfg = backbone.gnn_cfg
        self.reg_cfg = backbone.reg_cfg

        self.regress_forces = backbone.regress_forces
        self.direct_forces = backbone.direct_forces

    def post_init(self, gain=1.0):
        # init weights
        self.apply(partial(init_linear_weights, gain=gain))

    @torch.jit.ignore(drop=False)
    def no_weight_decay(self):
        return no_weight_decay(self)


@registry.register_model("EScAIP_direct_force_head")
class EScAIPDirectForceHead(EScAIPHeadBase):
    def __init__(self, backbone: EScAIPBackbone):  # type: ignore
        super().__init__(backbone)
        self.force_direction_layer = OutputLayer(
            global_cfg=self.global_cfg,
            gnn_cfg=self.gnn_cfg,
            reg_cfg=self.reg_cfg,
            output_type="Vector",
        )
        self.force_magnitude_layer = OutputLayer(
            global_cfg=self.global_cfg,
            gnn_cfg=self.gnn_cfg,
            reg_cfg=self.reg_cfg,
            output_type="Scalar",
        )
        self.node_norm = get_normalization_layer(self.reg_cfg.normalization)(
            self.global_cfg.hidden_size
        )
        self.edge_norm = get_normalization_layer(self.reg_cfg.normalization)(
            self.global_cfg.hidden_size
        )

        self.post_init()

    def compiled_forward(self, edge_features, node_features, data: GraphAttentionData):
        edge_features = self.edge_norm(edge_features)
        node_features = self.node_norm(node_features)

        # get force direction from edge features
        force_direction = self.force_direction_layer(
            edge_features
        )  # (num_nodes, max_neighbor, 3)
        force_direction = (
            force_direction * data.edge_direction
        )  # (num_nodes, max_neighbor, 3)
        force_direction = (force_direction * data.neighbor_mask.unsqueeze(-1)).sum(
            dim=1
        )  # (num_nodes, 3)
        # get force magnitude from node readouts
        force_magnitude = self.force_magnitude_layer(node_features)  # (num_nodes, 1)
        # get output force
        return force_direction * force_magnitude  # (num_nodes, 3)

    @conditional_grad(torch.enable_grad())
    def forward(self, data, emb: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        self.forward_fn = (
            torch.compile(self.compiled_forward)  # type: ignore
            if self.global_cfg.use_compile
            else self.compiled_forward
        )

        force_output = self.forward_fn(  # type: ignore
            edge_features=emb["edge_features"],
            node_features=emb["node_features"],
            data=emb["data"],  # type: ignore
        )

        return unpad_results(
            results={"forces": force_output},
            node_padding_mask=emb["data"].node_padding_mask,  # type: ignore
            graph_padding_mask=emb["data"].graph_padding_mask,  # type: ignore
        )


@registry.register_model("EScAIP_energy_head")
class EScAIPEnergyHead(EScAIPHeadBase):
    def __init__(self, backbone: EScAIPBackbone):  # type: ignore
        super().__init__(backbone)
        self.energy_layer = OutputLayer(
            global_cfg=self.global_cfg,
            gnn_cfg=self.gnn_cfg,
            reg_cfg=self.reg_cfg,
            output_type="Scalar",
        )
        self.energy_reduce = self.gnn_cfg.energy_reduce
        self.use_global_readout = self.gnn_cfg.use_global_readout
        self.node_norm = get_normalization_layer(self.reg_cfg.normalization)(
            self.global_cfg.hidden_size
        )

        self.post_init()

    def compiled_forward(self, emb):
        if self.use_global_readout:
            return self.energy_layer(self.node_norm(emb["global_features"]))

        energy_output = self.energy_layer(self.node_norm(emb["node_features"]))

        # the following not compatible with torch.compile (grpah break)
        # energy_output = torch_scatter.scatter(energy_output, node_batch, dim=0, reduce="sum")

        energy_output = compilable_scatter(
            src=energy_output,
            index=emb["data"].node_batch,
            dim_size=emb["data"].graph_padding_mask.shape[0],
            dim=0,
            reduce=self.energy_reduce,
        )
        return energy_output.squeeze()

    @conditional_grad(torch.enable_grad())
    def forward(self, data, emb: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        self.forward_fn = (
            torch.compile(self.compiled_forward)  # type: ignore
            if self.global_cfg.use_compile
            else self.compiled_forward
        )

        energy_output = self.forward_fn(emb)  # type: ignore
        if len(energy_output.shape) == 0:
            energy_output = energy_output.unsqueeze(0)
        return unpad_results(
            results={"energy": energy_output},
            node_padding_mask=emb["data"].node_padding_mask,  # type: ignore
            graph_padding_mask=emb["data"].graph_padding_mask,  # type: ignore
        )


@registry.register_model("EScAIP_grad_energy_force_stress_head")
class EScAIPGradientEnergyForceStressHead(EScAIPEnergyHead):  # type: ignore
    """
    Do not support torch.compile
    """

    def __init__(
        self,
        backbone: EScAIPBackbone,  # type: ignore
        prefix: str | None = None,
        wrap_property: bool = True,
    ):
        super().__init__(backbone)
        self.regress_stress = self.global_cfg.regress_stress
        self.regress_forces = self.global_cfg.regress_forces
        self.prefix = prefix
        self.wrap_property = wrap_property

    @conditional_grad(torch.enable_grad())
    def forward(self, data, emb: dict[str, torch.Tensor]) -> dict[str, torch.Tensor]:
        if self.prefix:
            energy_key = f"{self.prefix}_energy"
            forces_key = f"{self.prefix}_forces"
            stress_key = f"{self.prefix}_stress"
        else:
            energy_key = "energy"
            forces_key = "forces"
            stress_key = "stress"

        outputs = {}
        if self.use_global_readout:
            energy_output = self.energy_layer(emb["global_features"])
        else:
            energy_output = self.energy_layer(emb["node_features"])

            # the following not compatible with torch.compile (grpah break)
            # energy_output = torch_scatter.scatter(energy_output, node_batch, dim=0, reduce="sum")

            energy_output = compilable_scatter(
                src=energy_output,
                index=emb["data"].node_batch,  # type: ignore
                dim_size=emb["data"].graph_padding_mask.shape[0],  # type: ignore
                dim=0,
                reduce=self.energy_reduce,
            ).squeeze()
            if len(energy_output.shape) == 0:
                energy_output = energy_output.unsqueeze(0)
        outputs[energy_key] = (
            {"energy": energy_output} if self.wrap_property else energy_output
        )

        if self.regress_stress:
            grads = torch.autograd.grad(
                [energy_output.sum()],
                [data["pos_original"], emb["displacement"]],
                create_graph=self.training,
            )

            forces = torch.neg(grads[0])
            virial = grads[1].view(-1, 3, 3)
            volume = torch.det(data["cell"]).abs().unsqueeze(-1)
            stress = virial / volume.view(-1, 1, 1)
            virial = torch.neg(virial)
            stress = stress.view(
                -1, 9
            )  # NOTE to work better with current Multi-task trainer
            outputs[forces_key] = {"forces": forces} if self.wrap_property else forces
            outputs[stress_key] = {"stress": stress} if self.wrap_property else stress
            data["cell"] = emb["orig_cell"]
        elif self.regress_forces:
            forces = (
                -1
                * torch.autograd.grad(
                    energy_output.sum(), data["pos"], create_graph=self.training
                )[0]
            )
            outputs[forces_key] = {"forces": forces} if self.wrap_property else forces

        return unpad_results(
            results=outputs,
            node_padding_mask=emb["data"].node_padding_mask,  # type: ignore
            graph_padding_mask=emb["data"].graph_padding_mask,  # type: ignore
        )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import bisect
import copy
import logging
import os
import warnings
from abc import ABC, abstractmethod
from functools import cache, partial
from glob import glob
from pathlib import Path
from typing import Any, Callable

import ase
import numpy as np
from tqdm import tqdm

from fairchem.core.common.registry import registry
from fairchem.core.datasets._utils import rename_data_object_keys
from fairchem.core.datasets.atomic_data import AtomicData
from fairchem.core.datasets.base_dataset import BaseDataset
from fairchem.core.modules.transforms import DataTransforms


def apply_one_tags(
    atoms: ase.Atoms, skip_if_nonzero: bool = True, skip_always: bool = False
):
    """
    This function will apply tags of 1 to an ASE atoms object.
    It is used as an atoms_transform in the datasets contained in this file.

    Certain models will treat atoms differently depending on their tags.
    For example, GemNet-OC by default will only compute triplet and quadruplet interactions
    for atoms with non-zero tags. This model throws an error if there are no tagged atoms.
    For this reason, the default behavior is to tag atoms in structures with no tags.

    args:
        skip_if_nonzero (bool): If at least one atom has a nonzero tag, do not tag any atoms

        skip_always (bool): Do not apply any tags. This arg exists so that this function can be disabled
                without needing to pass a callable (which is currently difficult to do with main.py)
    """
    if skip_always:
        return atoms

    if np.all(atoms.get_tags() == 0) or not skip_if_nonzero:
        atoms.set_tags(np.ones(len(atoms)))

    return atoms


class AseAtomsDataset(BaseDataset, ABC):
    """
    This is an abstract Dataset that includes helpful utilities for turning
    ASE atoms objects into OCP-usable data objects. This should not be instantiated directly
    as get_atoms and _load_dataset_get_ids are not implemented in this base class.

    Derived classes must add at least two things:
        self.get_atoms(id): a function that takes an identifier and returns a corresponding atoms object

        self._load_dataset_get_ids(config: dict): This function is responsible for any initialization/loads
            of the dataset. The values it returns will be accessed by index using self.get_atoms(idx).

    Identifiers need not be any particular type.
    """

    def __init__(
        self,
        config: dict,
        atoms_transform: Callable[[ase.Atoms, Any, ...], ase.Atoms] = apply_one_tags,
    ) -> None:
        super().__init__(config)

        a2g_args = config.get("a2g_args", {}) or {}
        self.a2g = partial(AtomicData.from_ase, **a2g_args)

        self.key_mapping = self.config.get("key_mapping", None)
        self.transforms = DataTransforms(self.config.get("transforms", {}))

        self.atoms_transform = atoms_transform

        if self.config.get("keep_in_memory", False):
            self.__getitem__ = cache(self.__getitem__)

        self.ids = self._load_dataset_get_ids(config)
        self.num_samples = len(self.ids)

        if len(self.ids) == 0:
            raise ValueError(
                rf"No valid ase data found! \n"
                f"Double check that the src path and/or glob search pattern gives ASE compatible data: {config['src']}"
            )

    def __getitem__(self, idx):
        # Handle slicing
        if isinstance(idx, slice):
            return [self[i] for i in range(*idx.indices(len(self)))]

        # Get atoms object via derived class method
        atoms = self.get_atoms(idx)

        # Transform atoms object
        if self.atoms_transform is not None:
            atoms = self.atoms_transform(
                atoms, **self.config.get("atoms_transform_args", {})
            )

        sid = atoms.info.get("sid", self.ids[idx])

        fid = atoms.info.get("fid", None)
        if fid is not None:
            sid = f"{sid}-frame-{fid}"

        data_object = self.a2g(atoms, sid=sid)

        # Transform data object
        data_object = self.transforms(data_object)

        if self.key_mapping is not None:
            data_object = rename_data_object_keys(data_object, self.key_mapping)

        if self.config.get("include_relaxed_energy", False):
            data_object.energy_relaxed = self.get_relaxed_energy(self.ids[idx])

        return data_object

    @abstractmethod
    def get_atoms(self, idx: int) -> ase.Atoms:
        # This function should return an ASE atoms object.
        raise NotImplementedError(
            "Returns an ASE atoms object. Derived classes should implement this function."
        )

    @abstractmethod
    def _load_dataset_get_ids(self, config):
        # This function should return a list of ids that can be used to index into the database
        raise NotImplementedError(
            "Every ASE dataset needs to declare a function to load the dataset and return a list of ids."
        )

    def get_relaxed_energy(self, identifier):
        raise NotImplementedError(
            "Reading relaxed energy from trajectory or file is not implemented with this dataset. "
            "If relaxed energies are saved with the atoms info dictionary, they can be used by passing the keys in "
            "the r_data_keys argument under a2g_args."
        )

    def get_metadata(self, attr, idx):
        # try the parent method
        metadata = super().get_metadata(attr, idx)
        if metadata is not None:
            return metadata
        # try to resolve it here
        if attr != "natoms":
            return None
        if isinstance(idx, (list, np.ndarray)):
            return np.array([self.get_metadata(attr, i) for i in idx])
        return len(self.get_atoms(idx))


@registry.register_dataset("ase_read")
class AseReadDataset(AseAtomsDataset):
    """
    This Dataset uses ase.io.read to load data from a directory on disk.
    This is intended for small-scale testing and demonstrations of OCP.
    Larger datasets are better served by the efficiency of other dataset types
    such as LMDB.

    For a full list of ASE-readable filetypes, see
    https://wiki.fysik.dtu.dk/ase/ase/io/io.html

    args:
        config (dict):
            src (str): The source folder that contains your ASE-readable files

            pattern (str): Filepath matching each file you want to read
                    ex. "*/POSCAR", "*.cif", "*.xyz"
                    search recursively with two wildcards: "**/POSCAR" or "**/*.cif"

            a2g_args (dict): Keyword arguments for fairchem.core.preprocessing.AtomsToGraphs()
                    default options will work for most users

                    If you are using this for a training dataset, set
                    "r_energy":True, "r_forces":True, and/or "r_stress":True as appropriate
                    In that case, energy/forces must be in the files you read (ex. OUTCAR)

            ase_read_args (dict): Keyword arguments for ase.io.read()

            keep_in_memory (bool): Store data in memory. This helps avoid random reads if you need
                    to iterate over a dataset many times (e.g. training for many epochs).
                    Not recommended for large datasets.

            include_relaxed_energy (bool): Include the relaxed energy in the resulting data object.
                    The relaxed structure is assumed to be the final structure in the file
                    (e.g. the last frame of a .traj).

            atoms_transform_args (dict): Additional keyword arguments for the atoms_transform callable

            transform_args (dict): Additional keyword arguments for the transform callable

            key_mapping (dict[str, str]): Dictionary specifying a mapping between the name of a property used
                in the model with the corresponding property as it was named in the dataset. Only need to use if
                the name is different.

        atoms_transform (callable, optional): Additional preprocessing function applied to the Atoms
                    object. Useful for applying tags, for example.
    """

    def _load_dataset_get_ids(self, config) -> list[Path]:
        self.ase_read_args = config.get("ase_read_args", {})

        if ":" in self.ase_read_args.get("index", ""):
            raise NotImplementedError(
                "To read multiple structures from a single file, please use AseReadMultiStructureDataset."
            )

        self.path = Path(config["src"])
        if self.path.is_file():
            raise ValueError(
                f"The specified src is not a directory: {self.config['src']}"
            )

        if self.config.get("include_relaxed_energy", False):
            self.relaxed_ase_read_args = copy.deepcopy(self.ase_read_args)
            self.relaxed_ase_read_args["index"] = "-1"

        return list(self.path.glob(f'{config.get("pattern", "*")}'))

    def get_atoms(self, idx: int) -> ase.Atoms:
        try:
            file_path = self.ids[idx]
            atoms = ase.io.read(file_path, **self.ase_read_args)
        except Exception as err:
            warnings.warn(f"{err} occured for: {idx}", stacklevel=2)
            raise err

        if "sid" not in atoms.info:
            atoms.info["sid"] = str(file_path)

        return atoms

    def get_relaxed_energy(self, identifier) -> float:
        relaxed_atoms = ase.io.read(identifier, **self.relaxed_ase_read_args)
        return relaxed_atoms.get_potential_energy(apply_constraint=False)


@registry.register_dataset("ase_read_multi")
class AseReadMultiStructureDataset(AseAtomsDataset):
    """
    This Dataset can read multiple structures from each file using ase.io.read.
    The disadvantage is that all files must be read at startup.
    This is a significant cost for large datasets.

    This is intended for small-scale testing and demonstrations of OCP.
    Larger datasets are better served by the efficiency of other dataset types
    such as LMDB.

    For a full list of ASE-readable filetypes, see
    https://wiki.fysik.dtu.dk/ase/ase/io/io.html

    args:
        config (dict):
            src (str): The source folder that contains your ASE-readable files

            pattern (str): Filepath matching each file you want to read
                    ex. "*.traj", "*.xyz"
                    search recursively with two wildcards: "**/POSCAR" or "**/*.cif"

            index_file (str): Filepath to an indexing file, which contains each filename
                    and the number of structures contained in each file. For instance:

                    /path/to/relaxation1.traj 200
                    /path/to/relaxation2.traj 150

                    This will overrule the src and pattern that you specify!

            a2g_args (dict): Keyword arguments for fairchem.core.preprocessing.AtomsToGraphs()
                    default options will work for most users

                    If you are using this for a training dataset, set
                    "r_energy":True, "r_forces":True, and/or "r_stress":True as appropriate
                    In that case, energy/forces must be in the files you read (ex. OUTCAR)

            ase_read_args (dict): Keyword arguments for ase.io.read()

            keep_in_memory (bool): Store data in memory. This helps avoid random reads if you need
                    to iterate over a dataset many times (e.g. training for many epochs).
                    Not recommended for large datasets.

            include_relaxed_energy (bool): Include the relaxed energy in the resulting data object.
                    The relaxed structure is assumed to be the final structure in the file
                    (e.g. the last frame of a .traj).

            use_tqdm (bool): Use TQDM progress bar when initializing dataset

            atoms_transform_args (dict): Additional keyword arguments for the atoms_transform callable

            transform_args (dict): Additional keyword arguments for the transform callable

            key_mapping (dict[str, str]): Dictionary specifying a mapping between the name of a property used
                in the model with the corresponding property as it was named in the dataset. Only need to use if
                the name is different.

        atoms_transform (callable, optional): Additional preprocessing function applied to the Atoms
            object. Useful for applying tags, for example.

        transform (callable, optional): Additional preprocessing function for the Data object
    """

    def _load_dataset_get_ids(self, config) -> list[str]:
        self.ase_read_args = config.get("ase_read_args", {})
        if not hasattr(self.ase_read_args, "index"):
            self.ase_read_args["index"] = ":"

        if config.get("index_file", None) is not None:
            with open(config["index_file"]) as f:
                index = f.readlines()

            ids = []
            for line in index:
                filename = line.split(" ", maxsplit=1)[0]
                ids = [f"{filename} {i}" for i in range(int(line.split(" ")[1]))]

            return ids

        self.path = Path(config["src"])
        if self.path.is_file():
            raise ValueError(
                f"The specified src is not a directory: {self.config['src']}"
            )

        filenames = list(self.path.glob(f'{config.get("pattern", "*")}'))

        ids = []

        if config.get("use_tqdm", True):
            filenames = tqdm(filenames)
        for filename in filenames:
            try:
                structures = ase.io.read(filename, **self.ase_read_args)
            except Exception as err:
                warnings.warn(f"{err} occured for: {filename}", stacklevel=2)
            else:
                for i, _ in enumerate(structures):
                    ids.append(f"{filename} {i}")

        return ids

    def get_atoms(self, idx: int) -> ase.Atoms:
        try:
            std_id = self.ids[idx]
            identifiers = std_id.split(" ")
            atoms = ase.io.read("".join(identifiers[:-1]), **self.ase_read_args)[
                int(identifiers[-1])
            ]
        except Exception as err:
            warnings.warn(f"{err} occured for: {idx}", stacklevel=2)
            raise err

        if "sid" not in atoms.info:
            atoms.info["sid"] = "".join(identifiers[:-1])
        if "fid" not in atoms.info:
            atoms.info["fid"] = int(identifiers[-1])

        return atoms

    def get_relaxed_energy(self, identifier) -> float:
        relaxed_atoms = ase.io.read(
            "".join(identifier.split(" ")[:-1]), **self.ase_read_args
        )[-1]
        return relaxed_atoms.get_potential_energy(apply_constraint=False)


@registry.register_dataset("ase_db")
class AseDBDataset(AseAtomsDataset):
    """
    This Dataset connects to an ASE Database, allowing the storage of atoms objects
    with a variety of backends including JSON, SQLite, and database server options.

    For more information, see:
    https://databases.fysik.dtu.dk/ase/ase/db/db.html

    args:
        config (dict):
            src (str): Either
                    - the path an ASE DB,
                    - the connection address of an ASE DB,
                    - a folder with multiple ASE DBs,
                    - a list of folders with ASE DBs
                    - a glob string to use to find ASE DBs, or
                    - a list of ASE db paths/addresses.
                    If a folder, every file will be attempted as an ASE DB, and warnings
                    are raised for any files that can't connect cleanly

                    Note that for large datasets, ID loading can be slow and there can be many
                    ids, so it's advised to make loading the id list as easy as possible. There is not
                    an obvious way to get a full list of ids from most ASE dbs besides simply looping
                    through the entire dataset. See the AseLMDBDataset which was written with this usecase
                    in mind.

            connect_args (dict): Keyword arguments for ase.db.connect()

            select_args (dict): Keyword arguments for ase.db.select()
                    You can use this to query/filter your database

            a2g_args (dict): Keyword arguments for fairchem.core.preprocessing.AtomsToGraphs()
                    default options will work for most users

                    If you are using this for a training dataset, set
                    "r_energy":True, "r_forces":True, and/or "r_stress":True as appropriate
                    In that case, energy/forces must be in the database

            keep_in_memory (bool): Store data in memory. This helps avoid random reads if you need
                    to iterate over a dataset many times (e.g. training for many epochs).
                    Not recommended for large datasets.

            atoms_transform_args (dict): Additional keyword arguments for the atoms_transform callable

            transforms (dict[str, dict]): Dictionary specifying data transforms as {transform_function: config}
                    where config is a dictionary specifying arguments to the transform_function

            key_mapping (dict[str, str]): Dictionary specifying a mapping between the name of a property used
                in the model with the corresponding property as it was named in the dataset. Only need to use if
                the name is different.

        atoms_transform (callable, optional): Additional preprocessing function applied to the Atoms
                    object. Useful for applying tags, for example.

        transform (callable, optional): deprecated?
    """

    def _load_dataset_get_ids(self, config: dict) -> list[int]:
        if isinstance(config["src"], list):
            filepaths = []
            for path in sorted(config["src"]):
                if os.path.isdir(path):
                    filepaths.extend(sorted(glob(f"{path}/*")))
                elif os.path.isfile(path):
                    filepaths.append(path)
                else:
                    raise RuntimeError(f"Error reading dataset in {path}!")
        elif os.path.isfile(config["src"]):
            filepaths = [config["src"]]
        elif os.path.isdir(config["src"]):
            filepaths = sorted(glob(f'{config["src"]}/*'))
        else:
            filepaths = sorted(glob(config["src"]))

        self.dbs = []

        for path in filepaths:
            try:
                self.dbs.append(
                    self.connect_db(
                        path,
                        config.get("connect_args", {}),
                    )
                )
            except ValueError:
                logging.debug(
                    f"Tried to connect to {path} but it's not an ASE database!"
                )

        self.select_args = config.get("select_args", {})
        if self.select_args is None:
            self.select_args = {}

        # In order to get all of the unique IDs using the default ASE db interface
        # we have to load all the data and check ids using a select. This is extremely
        # inefficient for large dataset. If the db we're using already presents a list of
        # ids and there is no query, we can just use that list instead and save ourselves
        # a lot of time!
        self.db_ids = []
        for db in self.dbs:
            if hasattr(db, "ids") and self.select_args == {}:
                self.db_ids.append(db.ids)
            else:
                # this is the slow alternative
                self.db_ids.append([row.id for row in db.select(**self.select_args)])

        idlens = [len(ids) for ids in self.db_ids]
        self._idlen_cumulative = np.cumsum(idlens).tolist()

        return list(range(sum(idlens)))

    def get_atoms(self, idx: int) -> ase.Atoms:
        """Get atoms object corresponding to datapoint idx. Useful to read other properties not in data object.

        NOTE: if the row data in the database does not include a entry "sid" for the system id,
            The integer idx will be used.

        Args:
            idx (int): index in dataset

        Returns:
            atoms: ASE atoms corresponding to datapoint idx
        """
        # Figure out which db this should be indexed from.
        db_idx = bisect.bisect(self._idlen_cumulative, idx)

        # Extract index of element within that db
        el_idx = idx
        if db_idx != 0:
            el_idx = idx - self._idlen_cumulative[db_idx - 1]
        assert el_idx >= 0

        atoms_row = self.dbs[db_idx]._get_row(self.db_ids[db_idx][el_idx])
        atoms = atoms_row.toatoms()

        # put data back into atoms info
        if isinstance(atoms_row.data, dict):
            atoms.info.update(atoms_row.data)

        if "sid" not in atoms.info:
            atoms.info["sid"] = idx

        return atoms

    @staticmethod
    def connect_db(
        address: str | Path, connect_args: dict | None = None
    ) -> ase.db.core.Database:
        if connect_args is None:
            connect_args = {}

        # If we're using an aselmdb, let's set readonly=True to be safe!
        if "aselmdb" in address:
            connect_args["readonly"] = True
            connect_args["use_lock_file"] = False

        return ase.db.connect(address, **connect_args)

    def __del__(self):
        for db in self.dbs:
            if hasattr(db, "close"):
                db.close()



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import copy
from typing import Literal

import torch
import torch.nn as nn

from .radial import RadialMLP


class EdgeDegreeEmbedding(torch.nn.Module):
    """

    Args:
        sphere_channels (int):      Number of spherical channels

        lmax (int):                 degrees (l)
        mmax (int):                 orders (m)

        max_num_elements (int):     Maximum number of atomic numbers
        edge_channels_list (list:int):  List of sizes of invariant edge embedding. For example, [input_channels, hidden_channels, hidden_channels].
                                        The last one will be used as hidden size when `use_atom_edge_embedding` is `True`.
        use_atom_edge_embedding (bool): Whether to use atomic embedding along with relative distance for edge scalar features

        rescale_factor (float):     Rescale the sum aggregation
        cutoff (float):             Cutoff distance for the radial function

        mappingReduced (CoefficientMapping): Class to convert l and m indices once node embedding is rotated
    """

    def __init__(
        self,
        sphere_channels: int,
        lmax: int,
        mmax: int,
        edge_channels_list,
        rescale_factor,
        mappingReduced,
        # Enables activation checkpointing in size of
        # activation_checkpoint_chunk_size edge blocks
        activation_checkpoint_chunk_size: int | None,
    ):
        super().__init__()
        self.sphere_channels = sphere_channels
        self.lmax = lmax
        self.mmax = mmax
        self.mappingReduced = mappingReduced
        self.activation_checkpoint_chunk_size = activation_checkpoint_chunk_size

        self.m_0_num_coefficients: int = self.mappingReduced.m_size[0]
        self.m_all_num_coefficents: int = len(self.mappingReduced.l_harmonic)

        # Create edge scalar (invariant to rotations) features
        # Embedding function of the atomic numbers
        edge_channels_list = copy.deepcopy(edge_channels_list)

        # Embedding function of distance
        edge_channels_list.append(self.m_0_num_coefficients * self.sphere_channels)
        self.rad_func = RadialMLP(edge_channels_list)

        self.rescale_factor = rescale_factor

    def forward_chunk(
        self,
        x,
        x_edge,
        edge_index,
        wigner_and_M_mapping_inv,
        edge_envelope,
        node_offset=0,
    ):
        x_edge_m_0 = self.rad_func(x_edge)
        x_edge_m_0 = x_edge_m_0.reshape(
            -1, self.m_0_num_coefficients, self.sphere_channels
        )
        x_edge_embedding = torch.nn.functional.pad(
            x_edge_m_0,
            (0, 0, 0, (self.m_all_num_coefficents - self.m_0_num_coefficients)),
        )
        x_edge_embedding = torch.bmm(wigner_and_M_mapping_inv, x_edge_embedding)

        x_edge_embedding = x_edge_embedding * edge_envelope

        # TODO is this needed?
        x_edge_embedding = x_edge_embedding.to(x.dtype)

        return x.index_add(
            0, edge_index[1] - node_offset, x_edge_embedding / self.rescale_factor
        )

    def forward(
        self,
        x,
        x_edge,
        edge_index,
        wigner_and_M_mapping_inv,
        edge_envelope,
        node_offset=0,
    ):
        if self.activation_checkpoint_chunk_size is None:
            return self.forward_chunk(
                x,
                x_edge,
                edge_index,
                wigner_and_M_mapping_inv,
                edge_envelope,
                node_offset,
            )

        edge_index_partitions = edge_index.split(
            self.activation_checkpoint_chunk_size, dim=1
        )
        wigner_inv_partitions = wigner_and_M_mapping_inv.split(
            self.activation_checkpoint_chunk_size, dim=0
        )
        edge_envelope_partitions = edge_envelope.split(
            self.activation_checkpoint_chunk_size, dim=0
        )
        x_edge_partitions = x_edge.split(self.activation_checkpoint_chunk_size, dim=0)

        for idx in range(len(edge_index_partitions)):
            x = torch.utils.checkpoint.checkpoint(
                self.forward_chunk,
                x,
                x_edge_partitions[idx],
                edge_index_partitions[idx],
                wigner_inv_partitions[idx],
                edge_envelope_partitions[idx],
                node_offset,
                use_reentrant=False,
            )

        return x


class ChgSpinEmbedding(nn.Module):
    def __init__(
        self,
        embedding_type: Literal["pos_emb", "lin_emb", "rand_emb"],
        embedding_target: Literal["charge", "spin"],
        embedding_size: int,
        grad: bool,
        scale: float = 1.0,
    ) -> None:
        super().__init__()
        assert embedding_type in ["pos_emb", "lin_emb", "rand_emb"]
        self.embedding_type = embedding_type
        assert embedding_target in ["charge", "spin"]
        self.embedding_target = embedding_target
        assert embedding_size % 2 == 0, f"{embedding_size=} must be even"

        if self.embedding_target == "charge":
            # 100 is a conservative upper bound
            self.target_dict = {str(x): x + 100 for x in range(-100, 101)}
        elif self.embedding_target == "spin":
            # 100 is a conservative upper bound
            self.target_dict = {str(x): x for x in range(101)}

        if self.embedding_type == "pos_emb":
            # dividing by 2 because x_proj multiplies by 2
            if not grad:
                self.W = nn.Parameter(
                    torch.randn(embedding_size // 2) * scale, requires_grad=False
                )
            else:
                self.W = nn.Parameter(
                    torch.randn(embedding_size // 2) * scale, requires_grad=True
                )
        elif self.embedding_type == "lin_emb":
            self.lin_emb = nn.Linear(in_features=1, out_features=embedding_size)
            if not grad:
                for param in self.lin_emb.parameters():
                    param.requires_grad = False
        elif self.embedding_type == "rand_emb":
            self.rand_emb = nn.Embedding(len(self.target_dict), embedding_size)
            if not grad:
                for param in self.rand_emb.parameters():
                    param.requires_grad = False

        else:
            raise ValueError(f"embedding type {self.embedding_type} not implemented")

    def forward(self, x):
        # null token for spin is 0
        # charge is default 0
        if self.embedding_type == "pos_emb":
            x_proj = x[:, None] * self.W[None, :] * 2 * torch.pi
            if self.embedding_target == "charge":
                return torch.cat([torch.sin(x_proj), torch.cos(x_proj)], dim=-1)
            elif self.embedding_target == "spin":
                zero_idxs = torch.where(x == 0)[0]
                emb = torch.cat([torch.sin(x_proj), torch.cos(x_proj)], dim=-1)
                # this sets the null spin embedding to zero
                emb[zero_idxs] = 0
                return emb
        elif self.embedding_type == "lin_emb":
            if self.embedding_target == "spin":
                x[x == 0] = -100
            return self.lin_emb(x.unsqueeze(-1).float())
        elif self.embedding_type == "rand_emb":
            return self.rand_emb(
                torch.tensor(
                    [self.target_dict[str(i)] for i in x.tolist()],
                    device=x.device,
                    dtype=torch.long,
                )
            )
        raise ValueError(f"embedding type {self.embedding_type} not implemented")


class DatasetEmbedding(nn.Module):
    def __init__(self, embedding_size, grad, dataset_list):
        super().__init__()
        self.embedding_size = embedding_size
        self.dataset_emb_dict = nn.ModuleDict({})
        for dataset in dataset_list:
            if dataset not in self.dataset_emb_dict:
                self.dataset_emb_dict[dataset] = nn.Embedding(1, embedding_size)
            if not grad:
                for param in self.dataset_emb_dict[dataset].parameters():
                    param.requires_grad = False

    def forward(self, dataset_list):
        device = list(self.parameters())[0].device
        emb_idx = torch.tensor(0, device=device, dtype=torch.long)

        # TODO: this is a hack to accomodate the MPA finetuning
        # emb_for_datasets = [
        #     self.dataset_emb_dict[dataset](emb_idx) for dataset in dataset_list
        # ]
        emb_for_datasets = [
            self.dataset_emb_dict["omat"](emb_idx)
            if dataset in ["mptrj", "salex"]
            else self.dataset_emb_dict[dataset](emb_idx)
            for dataset in dataset_list
        ]

        return torch.stack(emb_for_datasets, dim=0)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the LICENSE
file in the root directory of this source tree.
"""

from __future__ import annotations

from dataclasses import dataclass

from fairchem.core.common.utils import StrEnum


class UMATask(StrEnum):
    OMOL = "omol"
    OMAT = "omat"
    ODAC = "odac"
    OC20 = "oc20"
    OC25 = "oc25"
    OMC = "omc"


CHARGE_RANGE = [-100, 100]
SPIN_RANGE = [0, 100]
DEFAULT_CHARGE = 0
DEFAULT_SPIN_OMOL = 1
DEFAULT_SPIN = 0


@dataclass
class MLIPInferenceCheckpoint:
    # contains original config that trained the model
    model_config: dict

    # the model state dict
    model_state_dict: dict

    # the ema state dict, used for inference
    ema_state_dict: dict

    # the config containing information about "tasks", a task contains
    # things like normalizers and element references and tells the model
    # how to produce the correct outputs
    tasks_config: dict


@dataclass
class InferenceSettings:
    # Flag to enable or disable the use of tf32 data type for inference.
    # TF32 will slightly reduce accuracy compared to FP32 but will still
    # keep energy conservation in most cases.
    tf32: bool = False

    # Flag to enable or disable activation checkpointing during
    # inference. This will dramatically decrease the memory footprint
    # especially for large number of atoms (ie 10+) at a slight cost to
    # inference speed. If set to None, the setting from the model
    # checkpoint will be used.
    activation_checkpointing: bool | None = None

    # Flag to enable or disable the merging of MOLE experts during
    # inference. If this is used, the input composition, total charge
    # and spin MUST remain constant throughout the simulation this will
    # slightly increase speed and reduce memory footprint used by the
    # parameters significantly
    merge_mole: bool = False

    # Flag to enable or disable the compilation of the inference model.
    compile: bool = False

    # Flag to enable or disable the use of CUDA Graphs for compute
    # This flag is no longer used and will be removed in future versions
    wigner_cuda: bool | None = None

    # Flag to enable or disable the generation of external graphs during
    # inference. If set to None, the setting from the model checkpoint
    # will be used.
    external_graph_gen: bool | None = None

    # Deprecated
    # Not recommended using! manually selects the version of graph gen
    # code if external_graph_gen is false, if set of None, will default
    # to whatever is in the checkpoint
    internal_graph_gen_version: int | None = None

    # Number of internal torch threads to use for inference
    torch_num_threads: int | None = None

    edge_chunk_size: int | None = None


# this is most general setting that works for most systems and models,
# not optimized for speed
def inference_settings_default():
    return InferenceSettings(
        tf32=False,
        activation_checkpointing=True,
        merge_mole=False,
        compile=False,
        external_graph_gen=False,
        internal_graph_gen_version=2,
    )


# this setting is designed for running long simulations or optimizations
# where the system composition (atoms, charge, spin) stays constant over
# the course the simulation. For smaller systems
# activation_checkpointing can be turned off for some extra speed gain
def inference_settings_turbo():
    return InferenceSettings(
        tf32=True,
        activation_checkpointing=False,
        merge_mole=True,
        compile=True,
        external_graph_gen=False,
        internal_graph_gen_version=2,
    )


# this mode corresponds to the default settings used for training and evaluation
def inference_settings_traineval():
    return InferenceSettings(
        tf32=False,
        activation_checkpointing=False,
        merge_mole=False,
        compile=False,
        internal_graph_gen_version=1,
    )


NAME_TO_INFERENCE_SETTING = {
    "default": inference_settings_default(),
    "turbo": inference_settings_turbo(),
}


def guess_inference_settings(settings: str | InferenceSettings):
    if isinstance(settings, str):
        assert (
            settings in NAME_TO_INFERENCE_SETTING
        ), f"inference setting name must be one of {NAME_TO_INFERENCE_SETTING.keys()}"
        return NAME_TO_INFERENCE_SETTING[settings]
    elif isinstance(settings, InferenceSettings):
        return settings
    else:
        raise ValueError(
            f"InferenceSetting can be a str or InferenceSettings object, found {settings.__class__}"
        )



from __future__ import annotations

import math

import torch
import torch.nn.functional as F
from e3nn.o3._spherical_harmonics import _spherical_harmonics


def get_node_direction_expansion_neighbor(
    direction_vec: torch.Tensor, neighbor_mask: torch.Tensor, lmax: int
):
    """
    Calculate Bond-Orientational Order (BOO) for each node in the graph.
    Ref: Steinhardt, et al. "Bond-orientational order in liquids and glasses." Physical Review B 28.2 (1983): 784.
    Input:
        direction_vec: (num_nodes, num_neighbors, 3)
        neighbor_mask: (num_nodes, num_neighbors)
    Return:
        node_boo: (num_nodes, num_neighbors, lmax + 1)
    """
    # Convert mask to float and expand dimensions
    neighbor_mask = neighbor_mask.float().unsqueeze(-1)

    # Compute spherical harmonics with proper normalization
    edge_sh = _spherical_harmonics(
        lmax=lmax,
        x=direction_vec[:, :, 0],
        y=direction_vec[:, :, 1],
        z=direction_vec[:, :, 2],
    )

    # Normalize spherical harmonics by sqrt(2l+1) to improve numerical stability
    sh_index = torch.arange(lmax + 1, device=edge_sh.device)
    sh_index = torch.repeat_interleave(sh_index, 2 * sh_index + 1)
    edge_sh = edge_sh / torch.clamp(torch.sqrt(2 * sh_index + 1), min=1e-6).unsqueeze(
        0
    ).unsqueeze(0)

    # Compute masked spherical harmonics
    masked_sh = edge_sh * neighbor_mask

    # Compute mean over neighbors with proper normalization
    neighbor_count = neighbor_mask.sum(dim=1)
    neighbor_count = torch.clamp(neighbor_count, min=1)
    node_boo = masked_sh.sum(dim=1) / neighbor_count

    # Compute final BOO with proper normalization
    node_boo_squared = node_boo**2
    # node_boo = scatter(node_boo_squared, sh_index, dim=-1, reduce="sum").sqrt()
    node_boo = compilable_scatter(
        node_boo_squared, sh_index, dim_size=lmax + 1, dim=-1, reduce="sum"
    )
    node_boo = torch.clamp(node_boo, min=1e-6).sqrt()

    return node_boo


def map_sender_receiver_feature(sender_feature, receiver_feature, neighbor_list):
    """
    Map from node features to edge features.
    sender_feature, receiver_feature: (num_nodes, h)
    neighbor_list: (num_nodes, max_neighbors)
    return: sender_features, receiver_features (num_nodes, max_neighbors, h)
    """
    # sender feature
    sender_feature = sender_feature[neighbor_list.flatten()].view(
        neighbor_list.shape[0], neighbor_list.shape[1], -1
    )

    # receiver features
    receiver_feature = receiver_feature.unsqueeze(1).expand(
        -1, neighbor_list.shape[1], -1
    )

    return (sender_feature, receiver_feature)


def legendre_polynomials(x: torch.Tensor, lmax: int) -> torch.Tensor:
    """
    Compute Legendre polynomials P_0..P_{lmax} for each element in x,
    using the standard recursion:
      P_0(x) = 1
      P_1(x) = x
      (n+1)*P_{n+1}(x) = (2n+1)*x*P_n(x) - n*P_{n-1}(x)
    x can have any shape; output will have an extra dimension (lmax+1).
    """
    # x shape could be (N, max_neighbors, max_neighbors) in your case
    shape = x.shape  # e.g. (N, M, M)
    P = x.new_zeros(*shape, lmax + 1)  # => (N, M, M, lmax+1)

    # P_0
    P[..., 0] = 1.0

    if lmax >= 1:
        # P_1
        P[..., 1] = x

        for n in range(1, lmax):
            P[..., n + 1] = ((2 * n + 1) * x * P[..., n] - n * P[..., n - 1]) / (n + 1)

    return P


def get_compact_frequency_vectors(
    edge_direction: torch.Tensor, lmax: int, repeating_dimensions: torch.Tensor | list
):
    """
    Calculate a compact representation of frequency vectors.
    Args:
        edge_direction: (N, k, 3) normalized direction vectors
        lmax: maximum l value for spherical harmonics
        repeating_dimensions: (lmax+1,) tensor or list with repeat counts for each l value
    Returns:
        frequency_vectors: (N, k, sum_{l=0..lmax} rep_l * (2l+1))
        flat tensor containing the spherical harmonics matched to repeating dimensions
    """
    # Convert to Python list if tensor to ensure it's treated as a compile-time constant
    if isinstance(repeating_dimensions, torch.Tensor):
        repeat_dims = repeating_dimensions.tolist()
    else:
        repeat_dims = repeating_dimensions

    edge_direction = edge_direction.to(torch.float32)
    # (edge_direction: N, k, 3)
    harmonics = _spherical_harmonics(
        lmax, edge_direction[..., 0], edge_direction[..., 1], edge_direction[..., 2]
    )  # (N, k, (lmax + 1)**2)

    # Create list to hold components
    components = []
    curr_idx = 0

    # Process each l value based on repeating dimensions
    for _l in range(lmax + 1):
        # Get the (2l+1) components for this l value
        sh_dim = 2 * _l + 1
        curr_irrep = harmonics[:, :, curr_idx : curr_idx + sh_dim] / math.sqrt(sh_dim)

        # Get repeat count - use Python list instead of tensor access
        rep_count = repeat_dims[_l]

        # Only add component if rep_count > 0
        if rep_count > 0:
            # Create a component that will match with the expanded q and k
            # (N, k, 2l+1) -> (N, k, rep_count * (2l+1))
            component = curr_irrep.unsqueeze(2).expand(-1, -1, rep_count, -1)
            component = component.reshape(component.shape[0], component.shape[1], -1)

            # Add component to list
            components.append(component)

        # Update index
        curr_idx += sh_dim

    # Concatenate components if we have any, otherwise return empty tensor
    if components:
        return torch.cat(components, dim=-1)
    else:
        # Return empty tensor with proper shape if no components
        return torch.zeros(
            (edge_direction.shape[0], edge_direction.shape[1], 0),
            device=edge_direction.device,
        )


def get_attn_mask(
    edge_direction: torch.Tensor,
    neighbor_mask: torch.Tensor,
    num_heads: int,
    lmax: int,
    use_angle_embedding: str,
):
    """
    Args:
        edge_direction: (num_nodes, max_neighbors, 3)
        neighbor_mask: (num_nodes, max_neighbors)
        num_heads: number of attention heads
    """
    # create a mask for empty neighbors
    batch_size, max_neighbors = neighbor_mask.shape
    attn_mask = torch.zeros(
        batch_size, max_neighbors, max_neighbors, device=neighbor_mask.device
    )
    attn_mask = attn_mask.masked_fill(~neighbor_mask.unsqueeze(1), float("-inf"))

    # repeat the mask for each head
    attn_mask = (
        attn_mask.unsqueeze(1)
        .expand(batch_size, num_heads, max_neighbors, max_neighbors)
        .reshape(batch_size * num_heads, max_neighbors, max_neighbors)
    )

    # get the angle embeddings
    dot_product = torch.matmul(edge_direction, edge_direction.transpose(1, 2))
    dot_product = dot_product.clamp(-1.0, 1.0)
    if use_angle_embedding == "bias":
        angle_embedding = legendre_polynomials(dot_product, lmax)
    else:
        angle_embedding = (
            dot_product.unsqueeze(1)
            .expand(-1, num_heads, -1, -1)
            .reshape(batch_size * num_heads, max_neighbors, max_neighbors)
        )
    return attn_mask, angle_embedding


def get_attn_mask_env(
    src_mask: torch.Tensor,
    num_heads: int,
):
    """
    Args:
        src_mask: (num_nodes, num_neighbors)
        num_heads: number of attention heads
    Output:
        attn_mask: (num_nodes * num_heads, num_neighbors, num_neighbors)
    """
    batch_size, num_neighbors = src_mask.shape
    # broadcast src_mask to attention mask shape
    attn_mask = (
        src_mask.unsqueeze(1)
        .unsqueeze(2)  # (num_nodes, 1, 1, num_neighbors)
        .expand(
            -1, num_heads, num_neighbors, -1
        )  # (num_nodes, num_heads, num_neighbors, num_neighbors)
        .reshape(batch_size * num_heads, num_neighbors, num_neighbors)
    )
    return attn_mask


def pad_batch(
    max_atoms,
    max_batch_size,
    atomic_numbers,
    node_direction_expansion,
    edge_distance_expansion,
    edge_direction,
    neighbor_list,
    neighbor_mask,
    node_batch,
    num_graphs,
    src_mask=None,
):
    """
    Pad the batch to have the same number of nodes in total.
    Needed for torch.compile

    Note: the sampler for multi-node training could sample batchs with different number of graphs.
    The number of sampled graphs could be smaller or larger than the batch size.
    This would cause the model to recompile or core dump.
    Temporarily, setting the max number of graphs to be twice the batch size to mitigate this issue.
    TODO: look into a better way to handle this.
    """
    device = atomic_numbers.device
    num_nodes, _ = neighbor_list.shape
    pad_size = max_atoms - num_nodes
    assert (
        pad_size >= 0
    ), "Number of nodes exceeds the maximum number of nodes per batch"

    # pad the features
    atomic_numbers = F.pad(atomic_numbers, (0, pad_size), value=0)
    node_direction_expansion = F.pad(
        node_direction_expansion, (0, 0, 0, pad_size), value=0
    )
    edge_distance_expansion = F.pad(
        edge_distance_expansion, (0, 0, 0, 0, 0, pad_size), value=0
    )
    edge_direction = F.pad(edge_direction, (0, 0, 0, 0, 0, pad_size), value=0)
    neighbor_list = F.pad(neighbor_list, (0, 0, 0, pad_size), value=-1)
    neighbor_mask = F.pad(neighbor_mask, (0, 0, 0, pad_size), value=0)
    node_batch = F.pad(node_batch, (0, pad_size), value=num_graphs)
    if src_mask is not None:
        src_mask = F.pad(src_mask, (0, 0, 0, pad_size), value=0)

    # create the padding mask
    node_padding_mask = torch.ones(max_atoms, dtype=torch.bool, device=device)
    node_padding_mask[num_nodes:] = False

    # TODO look into a better way to handle this
    graph_padding_mask = torch.ones(max_batch_size, dtype=torch.bool, device=device)
    graph_padding_mask[num_graphs:] = False

    return (
        atomic_numbers,
        node_direction_expansion,
        edge_distance_expansion,
        edge_direction,
        neighbor_list,
        neighbor_mask,
        src_mask,
        node_batch,
        node_padding_mask,
        graph_padding_mask,
    )


def unpad_results(results, node_padding_mask, graph_padding_mask):
    """
    Unpad the results to remove the padding.
    """
    unpad_results = {}
    for key in results:
        if results[key].shape[0] == node_padding_mask.shape[0]:
            unpad_results[key] = results[key][node_padding_mask]
        elif results[key].shape[0] == graph_padding_mask.shape[0]:
            unpad_results[key] = results[key][graph_padding_mask]
        elif (
            results[key].shape[0] == node_padding_mask.sum()
            or results[key].shape[0] == graph_padding_mask.sum()
        ):
            unpad_results[key] = results[key]
        else:
            raise ValueError("Unknown padding mask shape")
    return unpad_results


def patch_singleton_atom(edge_direction, neighbor_list, neighbor_mask):
    """
    Patch the singleton atoms in the neighbor_list and neighbor_mask.
    Add a self-loop to the singleton atom
    """

    # Find the singleton atoms
    idx = torch.where(neighbor_mask.sum(dim=-1) == 0)[0]

    # patch edge_direction to unit vector
    edge_direction[idx, 0] = torch.tensor(
        [1.0, 0.0, 0.0], device=edge_direction.device, dtype=edge_direction.dtype
    )

    # patch neighbor_list to itself
    neighbor_list[idx, 0] = idx

    # patch neighbor_mask to itself
    neighbor_mask[idx, 0] = 1

    return edge_direction, neighbor_list, neighbor_mask


def compilable_scatter(
    src: torch.Tensor,
    index: torch.Tensor,
    dim_size: int,
    dim: int = 0,
    reduce: str = "sum",
) -> torch.Tensor:
    """
    torch_scatter scatter function with compile support.
    Modified from torch_geometric.utils.scatter_.
    """

    def broadcast(src: torch.Tensor, ref: torch.Tensor, dim: int) -> torch.Tensor:
        dim = ref.dim() + dim if dim < 0 else dim
        size = ((1,) * dim) + (-1,) + ((1,) * (ref.dim() - dim - 1))
        return src.view(size).expand_as(ref)

    dim = src.dim() + dim if dim < 0 else dim
    size = src.size()[:dim] + (dim_size,) + src.size()[dim + 1 :]

    if reduce == "sum" or reduce == "add":
        index = broadcast(index, src, dim)
        return src.new_zeros(size).scatter_add_(dim, index, src)

    if reduce == "mean":
        count = src.new_zeros(dim_size)
        count.scatter_add_(0, index, src.new_ones(src.size(dim)))
        count = count.clamp(min=1)

        index = broadcast(index, src, dim)
        out = src.new_zeros(size).scatter_add_(dim, index, src)

        return out / broadcast(count, out, dim)

    raise ValueError(f"Invalid reduce option '{reduce}'.")


def get_displacement_and_cell(data, regress_stress, regress_forces, direct_forces):
    """
    Get the displacement and cell from the data.
    For gradient-based forces/stress
    ref: https://github.com/facebookresearch/fairchem/blob/main/src/fairchem/core/models/uma/escn_md.py#L298
    """
    displacement = None
    orig_cell = None
    if regress_stress and not direct_forces:
        displacement = torch.zeros(
            (3, 3),
            dtype=data["pos"].dtype,
            device=data["pos"].device,
        )
        num_batch = len(data["natoms"])
        displacement = displacement.view(-1, 3, 3).expand(num_batch, 3, 3)
        displacement.requires_grad = True
        symmetric_displacement = 0.5 * (displacement + displacement.transpose(-1, -2))
        if data["pos"].requires_grad is False:
            data["pos"].requires_grad = True
        data["pos_original"] = data["pos"]
        data["pos"] = data["pos"] + torch.bmm(
            data["pos"].unsqueeze(-2),
            torch.index_select(symmetric_displacement, 0, data["batch"]),
        ).squeeze(-2)

        orig_cell = data["cell"]
        data["cell"] = data["cell"] + torch.bmm(data["cell"], symmetric_displacement)
    if (
        not regress_stress
        and regress_forces
        and not direct_forces
        and data["pos"].requires_grad is False
    ):
        data["pos"].requires_grad = True
    return displacement, orig_cell



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import torch
import torch.nn.functional as F


class Act(torch.nn.Module):
    def __init__(self, act: str, slope: float = 0.05) -> None:
        super().__init__()
        self.act = act
        self.slope = slope
        self.shift = torch.log(torch.tensor(2.0)).item()

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        if self.act == "relu":
            return F.relu(input)
        elif self.act == "leaky_relu":
            return F.leaky_relu(input)
        elif self.act == "sp":
            return F.softplus(input, beta=1)
        elif self.act == "leaky_sp":
            return F.softplus(input, beta=1) - self.slope * F.relu(-input)
        elif self.act == "elu":
            return F.elu(input, alpha=1)
        elif self.act == "leaky_elu":
            return F.elu(input, alpha=1) - self.slope * F.relu(-input)
        elif self.act == "ssp":
            return F.softplus(input, beta=1) - self.shift
        elif self.act == "leaky_ssp":
            return F.softplus(input, beta=1) - self.slope * F.relu(-input) - self.shift
        elif self.act == "tanh":
            return torch.tanh(input)
        elif self.act == "leaky_tanh":
            return torch.tanh(input) + self.slope * input
        elif self.act == "swish":
            return torch.sigmoid(input) * input
        else:
            raise RuntimeError(f"Undefined activation called {self.act}")



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import logging
from typing import Literal

import torch
from torch import nn

from fairchem.core.common import distutils, gp_utils
from fairchem.core.common.registry import registry


class DDPMTLoss(nn.Module):
    """
    This class is a wrapper around a loss function that does a few things
    like handle nans and importantly ensures the reduction is done
    correctly for DDP. The main issue is that DDP averages gradients
    over replicas — this only works out of the box if the dimension
    you are averaging over is completely consistent across all replicas.
    In our case, that is not true for the number of atoms per batch and
    there are edge cases when the batch size differs between replicas
    e.g. if the dataset size is not divisible by the batch_size.

    Scalars are relatively straightforward to handle, but vectors and higher tensors
    are a bit trickier. Below are two examples of forces.

    Forces input: [Nx3] target: [Nx3]
    Forces are a vector of length 3 (x,y,z) for each atom.
    Number of atoms per batch (N) is different for each DDP replica.

    MSE example:
    #### Local loss computation ####
    local_loss = MSELoss(input, target) -> [Nx3]
    num_samples = local_loss.numel() -> [Nx3]
    local_loss = sum(local_loss [Nx3]) -> [1] sum reduces the loss to a scalar
    global_samples = all_reduce(num_samples) -> [N0x3 + N1x3 + N2x3 + ...] = [1] where N0 is the number of atoms on replica 0
    local_loss = local_loss * world_size / global_samples -> [1]
    #### Global loss computation ####
    global_loss = sum(local_loss / world_size) -> [1]
    == sum(local_loss / global_samples) # this is the desired corrected mean

    Norm example:
    #### Local loss computation ####
    local_loss = L2MAELoss(input, target) -> [N]
    num_samples = local_loss.numel() -> [N]
    local_loss = sum(local_loss [N]) -> [1] sum reduces the loss to a scalar
    global_samples = all_reduce(num_samples) -> [N0 + N1 + N2 + ...] = [1] where N0 is the number of atoms on replica 0
    local_loss = local_loss * world_size / global_samples -> [1]
    #### Global loss computation ####
    global_loss = sum(local_loss / world_size) -> [1]
    == sum(local_loss / global_samples) # this is the desired corrected mean
    """

    def __init__(
        self,
        loss_fn: torch.nn.Module,
        reduction: Literal["mean", "sum", "per_structure"] = "mean",
        coefficient: float = 1.0,
    ) -> None:
        super().__init__()
        self.loss_fn = loss_fn
        self.reduction = reduction
        self.reduction_map = {
            "mean": self.mean,
            "sum": self.sum,
            "per_structure": self.per_structure,
        }
        self.coefficient = coefficient
        assert self.reduction in list(
            self.reduction_map.keys()
        ), "Reduction must be one of: 'mean', 'sum', 'per_structure'"

    def sum(self, input, mult_mask, num_samples, loss, natoms):
        # this sum will reduce the loss down to a single scalar
        return torch.sum(loss)

    def _ddp_mean(self, num_samples, loss):
        # global_samples can be 0 if the head has no valid samples in the batch
        # protect against division by zero
        global_samples = max(distutils.all_reduce(num_samples, device=loss.device), 1)
        # Multiply by world size since gradients are averaged across DDP replicas
        # warning this is probably incorrect for any model parallel approach
        # Graph parallel note: numerator and denominator are inflated by the same
        # constant. # of processes in a single graph parallel group , which makes this
        # a strange way to implement the loss, but technically correct
        # however the gradient is not correct , please see comments at FixGPGrad()
        corrected_loss = loss * distutils.get_world_size() / global_samples

        if gp_utils.initialized():
            # make this explict so its easier to reason about loss here
            # calling fix_gp_grad in non-gp has no affect
            return gp_utils.scale_backward_grad(corrected_loss)
        return corrected_loss

    def mean(self, input, mult_mask, num_samples, loss, natoms):
        # this sum will reduce the loss down from num_sample -> 1
        loss = self.sum(input, mult_mask, num_samples, loss, natoms)
        return self._ddp_mean(num_samples, loss)

    def per_structure(self, input, mult_mask, num_samples, loss, natoms):
        struct_idx = torch.repeat_interleave(
            torch.arange(natoms.numel(), device=input.device), natoms
        )
        assert torch.unique(struct_idx).numel() == natoms.numel()
        per_struct_loss = torch.zeros(
            natoms.numel(), device=input.device
        ).scatter_reduce(0, struct_idx, loss, reduce="sum")

        # normalize by the number of free atoms in the structure
        free_natoms = torch.bincount(struct_idx[mult_mask], minlength=natoms.numel())
        zero_idx = torch.where(free_natoms == 0)[0]
        free_natoms[zero_idx] = natoms[zero_idx]
        assert torch.all(free_natoms > 0)
        assert torch.all(free_natoms <= natoms)
        per_struct_loss = per_struct_loss / free_natoms

        # takes the mean across all systems in the batch
        num_samples = torch.nonzero(per_struct_loss).numel()
        return self._ddp_mean(num_samples, per_struct_loss.sum())

    def _reduction(self, input, mult_mask, loss, natoms):
        num_samples = loss[mult_mask].numel()
        if self.reduction in self.reduction_map:
            return self.reduction_map[self.reduction](
                input, mult_mask, num_samples, loss, natoms
            )
        else:
            raise ValueError("Reduction must be one of: 'mean', 'sum'")

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        mult_mask: torch.Tensor,
        natoms: torch.Tensor,
    ):
        # ensure torch doesn't do any unwanted broadcasting
        assert (
            input.shape[0] == target.shape[0] == mult_mask.shape[0]
        ), f"Mismatched shapes: {input.shape} and {target.shape} and {mult_mask.shape}"

        # Ensure torch doesn't do any unwanted broadcasting
        target = target.view(input.shape)
        if input.numel() == mult_mask.numel():
            mult_mask = mult_mask.view(input.shape)

        loss = (
            self.loss_fn(
                input, torch.nan_to_num(target, posinf=0.0, neginf=0.0), natoms
            )
            * mult_mask
        )
        loss = self._reduction(input, mult_mask, loss, natoms)

        # Zero out nans, if any
        found_nans_or_infs = not torch.all(loss.isfinite())
        if found_nans_or_infs is True:
            logging.warning("Found nans while computing loss")
            loss = torch.nan_to_num(loss, nan=0.0)

        return self.coefficient * loss


@registry.register_loss("mae")
class MAELoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.loss = nn.L1Loss()
        # reduction should be none as it is handled in DDPLoss
        self.loss.reduction = "none"

    def forward(
        self, pred: torch.Tensor, target: torch.Tensor, natoms: torch.Tensor
    ) -> torch.Tensor:
        return self.loss(pred, target)


@registry.register_loss("mse")
class MSELoss(nn.Module):
    def __init__(self) -> None:
        super().__init__()
        self.loss = nn.MSELoss()
        # reduction should be none as it is handled in DDPLoss
        self.loss.reduction = "none"

    def forward(
        self, pred: torch.Tensor, target: torch.Tensor, natoms: torch.Tensor
    ) -> torch.Tensor:
        return self.loss(pred, target)


@registry.register_loss("per_atom_mae")
class PerAtomMAELoss(nn.Module):
    """
    Simply divide a loss by the number of atoms/nodes in the graph.
    Current this loss is intened to used with scalar values, not vectors or higher tensors.
    """

    def __init__(self) -> None:
        super().__init__()
        self.loss = nn.L1Loss()
        # reduction should be none as it is handled in DDPLoss
        self.loss.reduction = "none"

    def forward(
        self, pred: torch.Tensor, target: torch.Tensor, natoms: torch.Tensor
    ) -> torch.Tensor:
        _natoms = torch.reshape(natoms, target.shape)
        # check if target is a scalar
        assert target.dim() == 1 or (target.dim() == 2 and target.shape[1] == 1)
        # check per_atom shape
        assert (target / _natoms).shape == target.shape
        return self.loss(pred / _natoms, target / _natoms)


@registry.register_loss("l2norm")
@registry.register_loss("l2mae")
class L2NormLoss(nn.Module):
    """
    Currently this loss is intened to used with vectors.
    """

    def __init__(self) -> None:
        super().__init__()

    def forward(
        self, pred: torch.Tensor, target: torch.Tensor, natoms: torch.Tensor
    ) -> torch.Tensor:
        assert target.dim() == 2
        assert target.shape[1] != 1
        return torch.linalg.vector_norm(pred - target, ord=2, dim=-1)


class DDPLoss(nn.Module):
    """
    This class is a wrapper around a loss function that does a few things
    like handle nans and importantly ensures the reduction is done
    correctly for DDP. The main issue is that DDP averages gradients
    over replicas — this only works out of the box if the dimension
    you are averaging over is completely consistent across all replicas.
    In our case, that is not true for the number of atoms per batch and
    there are edge cases when the batch size differs between replicas
    e.g. if the dataset size is not divisible by the batch_size.

    Scalars are relatively straightforward to handle, but vectors and higher tensors
    are a bit trickier. Below are two examples of forces.

    Forces input: [Nx3] target: [Nx3]
    Forces are a vector of length 3 (x,y,z) for each atom.
    Number of atoms per batch (N) is different for each DDP replica.

    MSE example:
    #### Local loss computation ####
    local_loss = MSELoss(input, target) -> [Nx3]
    num_samples = local_loss.numel() -> [Nx3]
    local_loss = sum(local_loss [Nx3]) -> [1] sum reduces the loss to a scalar
    global_samples = all_reduce(num_samples) -> [N0x3 + N1x3 + N2x3 + ...] = [1] where N0 is the number of atoms on replica 0
    local_loss = local_loss * world_size / global_samples -> [1]
    #### Global loss computation ####
    global_loss = sum(local_loss / world_size) -> [1]
    == sum(local_loss / global_samples) # this is the desired corrected mean

    Norm example:
    #### Local loss computation ####
    local_loss = L2MAELoss(input, target) -> [N]
    num_samples = local_loss.numel() -> [N]
    local_loss = sum(local_loss [N]) -> [1] sum reduces the loss to a scalar
    global_samples = all_reduce(num_samples) -> [N0 + N1 + N2 + ...] = [1] where N0 is the number of atoms on replica 0
    local_loss = local_loss * world_size / global_samples -> [1]
    #### Global loss computation ####
    global_loss = sum(local_loss / world_size) -> [1]
    == sum(local_loss / global_samples) # this is the desired corrected mean
    """

    def __init__(
        self,
        loss_name,
        reduction: Literal["mean", "sum"],
    ) -> None:
        super().__init__()
        self.loss_fn = registry.get_loss_class(loss_name)()
        # default reduction is mean
        self.reduction = reduction if reduction is not None else "mean"
        self.reduction_map = {
            "mean": self.mean,
            "sum": self.sum,
        }
        assert self.reduction in list(
            self.reduction_map.keys()
        ), "Reduction must be one of: 'mean', 'sum'"

    def sum(self, input, loss, natoms):
        # this sum will reduce the loss down to a single scalar
        return torch.sum(loss)

    def _ddp_mean(self, num_samples, loss):
        global_samples = distutils.all_reduce(num_samples, device=loss.device)
        # Multiply by world size since gradients are averaged across DDP replicas
        # warning this is probably incorrect for any model parallel approach
        return loss * distutils.get_world_size() / global_samples

    def mean(self, input, loss, natoms):
        # total elements to take the mean over
        # could be batch_size, num_atoms, num_atomsx3, etc
        num_samples = loss.numel()
        # this sum will reduce the loss down from num_sample -> 1
        loss = self.sum(input, loss, natoms)
        return self._ddp_mean(num_samples, loss)

    def _reduction(self, input, loss, natoms):
        if self.reduction in self.reduction_map:
            return self.reduction_map[self.reduction](input, loss, natoms)
        else:
            raise ValueError("Reduction must be one of: 'mean', 'sum'")

    def forward(
        self,
        input: torch.Tensor,
        target: torch.Tensor,
        natoms: torch.Tensor,
    ):
        # ensure torch doesn't do any unwanted broadcasting
        assert (
            input.shape == target.shape
        ), f"Mismatched shapes: {input.shape} and {target.shape}"

        # zero out nans, if any
        found_nans_or_infs = not torch.all(input.isfinite())
        if found_nans_or_infs is True:
            logging.warning("Found nans while computing loss")
            input = torch.nan_to_num(input, nan=0.0)

        loss = self.loss_fn(input, target, natoms)
        return self._reduction(input, loss, natoms)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import functools
from contextlib import suppress

import torch
import torch.nn as nn

from fairchem.core.models.uma.nn.mole import (
    MOLE,
    MOLEDGL,
    MOLEGlobals,
    norm_str_to_fn,
)
from fairchem.core.models.uma.nn.so2_layers import SO2_Convolution

fairchem_cpp_found = False
with suppress(ModuleNotFoundError):
    fairchem_cpp_found = True


class MOLEInterface:
    def set_MOLE_coefficients(
        self, atomic_numbers_full, batch_full, csd_mixed_emb
    ) -> None:
        return None

    def set_MOLE_sizes(self, nsystems, batch_full, edge_index) -> None:
        return None

    def log_MOLE_stats(self) -> None:
        return None

    def merge_MOLE_model(self, data):
        return self


def recursive_replace_so2m0_linear(model, replacement_factory):
    for _, child in model.named_children():
        if isinstance(child, torch.nn.Module):
            recursive_replace_so2m0_linear(child, replacement_factory)
        if isinstance(child, SO2_Convolution):
            target_device = child.fc_m0.weight.device
            child.fc_m0 = replacement_factory(child.fc_m0).to(target_device)


def recursive_replace_so2_MOLE(model, replacement_factory):
    for _, child in model.named_children():
        if isinstance(child, torch.nn.Module):
            recursive_replace_so2_MOLE(child, replacement_factory)
        if isinstance(child, SO2_Convolution):
            target_device = child.fc_m0.weights.device
            child.fc_m0 = replacement_factory(child.fc_m0).to(target_device)
            for so2_module in child.so2_m_conv:
                so2_module.fc = replacement_factory(so2_module.fc).to(target_device)


def recursive_replace_so2_linear(model, replacement_factory):
    for _, child in model.named_children():
        if isinstance(child, torch.nn.Module):
            recursive_replace_so2_linear(child, replacement_factory)
        if isinstance(child, SO2_Convolution):
            target_device = child.fc_m0.weight.device
            child.fc_m0 = replacement_factory(child.fc_m0).to(target_device)
            for so2_module in child.so2_m_conv:
                so2_module.fc = replacement_factory(so2_module.fc).to(target_device)


def recursive_replace_all_linear(model, replacement_factory):
    for child_name, child in model.named_children():
        if isinstance(child, torch.nn.Linear):
            target_device = child.weight.device
            setattr(model, child_name, replacement_factory(child).to(target_device))
        elif isinstance(child, torch.nn.Module):
            recursive_replace_all_linear(child, replacement_factory)


def recursive_replace_notso2_linear(model, replacement_factory):
    for child_name, child in model.named_children():
        if isinstance(child, SO2_Convolution):
            continue
        if isinstance(child, torch.nn.Linear):
            target_device = child.weight.device
            setattr(model, child_name, replacement_factory(child).to(target_device))
        elif isinstance(child, torch.nn.Module):
            recursive_replace_notso2_linear(child, replacement_factory)


def model_search_and_replace(
    model, module_search_function, replacement_factory, layers=None
):
    if layers is None:
        layers = list(range(len(model.blocks)))
    for layer_idx in layers:
        module_search_function(model.blocks[layer_idx], replacement_factory)


def replace_linear_with_shared_linear(
    existing_linear_module,
    cache,
):
    layer_identifier = (
        existing_linear_module.in_features,
        existing_linear_module.out_features,
        existing_linear_module.bias is not None,
    )
    if layer_identifier in cache:
        return cache[layer_identifier]

    cache[layer_identifier] = existing_linear_module
    return existing_linear_module


def replace_MOLE_with_linear(
    existing_mole_module: MOLE,
):
    return existing_mole_module.merged_linear_layer()


def replace_linear_with_MOLE(
    existing_linear_module,
    global_mole_tensors,
    num_experts,
    mole_layer_type,
    cache=None,
):
    layer_identifier = (
        existing_linear_module.in_features,
        existing_linear_module.out_features,
        existing_linear_module.bias,
    )
    if cache is not None and layer_identifier in cache:
        return cache[layer_identifier]

    if mole_layer_type == "dgl":
        assert (
            fairchem_cpp_found
        ), "Cannot use DGL layer type if fairchem_cpp package is not available"
        layer = MOLEDGL(
            num_experts=num_experts,
            global_mole_tensors=global_mole_tensors,
            in_features=existing_linear_module.in_features,
            out_features=existing_linear_module.out_features,
            bias=existing_linear_module.bias is not None,
        )
    elif mole_layer_type == "pytorch":
        layer = MOLE(
            num_experts=num_experts,
            global_mole_tensors=global_mole_tensors,
            in_features=existing_linear_module.in_features,
            out_features=existing_linear_module.out_features,
            bias=existing_linear_module.bias is not None,
        )
    else:
        raise ValueError("mole_layer_type must be pytorch")
    if cache is not None:
        cache[layer_identifier] = layer
    return layer


def convert_model_to_MOLE_model(
    model,
    num_experts: int = 8,
    mole_dropout: float = 0.0,
    mole_expert_coefficient_norm: str = "softmax",
    act=torch.nn.SiLU,
    layers_mole=None,
    use_composition_embedding: bool = False,
    mole_layer_type: str = "pytorch",
    mole_single: bool = False,
    mole_type: str = "so2",
):
    model.num_experts = num_experts
    if model.num_experts == 0:
        return

    model.mole_type = mole_type

    routing_mlp_dim = (
        use_composition_embedding + 1  # always use dataset/csd_mixed_emb
    ) * model.sphere_channels

    model.routing_mlp = nn.Sequential(
        nn.Linear(
            routing_mlp_dim,
            num_experts * 2,
            bias=True,
        ),
        nn.SiLU(),
        nn.Linear(
            num_experts * 2,
            num_experts * 2,
            bias=True,
        ),
        nn.SiLU(),
        nn.Linear(
            num_experts * 2,
            num_experts,
            bias=True,
        ),
        nn.SiLU(),
    )

    #
    model.use_composition_embedding = use_composition_embedding
    model.global_mole_tensors = MOLEGlobals(
        expert_mixing_coefficients=None, mole_sizes=None
    )

    model.mole_dropout = torch.nn.Dropout(mole_dropout)
    model.mole_expert_coefficient_norm = norm_str_to_fn(mole_expert_coefficient_norm)
    model.act = act()

    if model.use_composition_embedding:
        model.composition_embedding = nn.Embedding(
            model.max_num_elements, model.sphere_channels
        )

    # plotting
    model.counter = 0

    # replace modules
    replacement_factory = functools.partial(
        replace_linear_with_MOLE,
        num_experts=model.num_experts,
        global_mole_tensors=model.global_mole_tensors,
        mole_layer_type=mole_layer_type,
        cache={} if mole_single else None,
    )

    if mole_type == "so2":
        model_search_and_replace(
            model, recursive_replace_so2_linear, replacement_factory, layers=layers_mole
        )
    elif mole_type == "so2m0":
        model_search_and_replace(
            model,
            recursive_replace_so2m0_linear,
            replacement_factory,
            layers=layers_mole,
        )
    elif mole_type == "all":
        model_search_and_replace(
            model, recursive_replace_all_linear, replacement_factory, layers=layers_mole
        )
    elif mole_type == "notso2":
        model_search_and_replace(
            model,
            recursive_replace_notso2_linear,
            replacement_factory,
            layers=layers_mole,
        )
    else:
        raise ValueError(f"Not a valid mole_type {mole_type}")



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import math

import torch


### Methods for sample points on a sphere
def CalcSpherePoints(num_points: int, device: str = "cpu") -> torch.Tensor:
    goldenRatio = (1 + 5**0.5) / 2
    i = torch.arange(num_points, device=device).view(-1, 1)
    theta = 2 * math.pi * i / goldenRatio
    phi = torch.arccos(1 - 2 * (i + 0.5) / num_points)
    points = torch.cat(
        [
            torch.cos(theta) * torch.sin(phi),
            torch.sin(theta) * torch.sin(phi),
            torch.cos(phi),
        ],
        dim=1,
    )

    # weight the points by their density
    pt_cross = points.view(1, -1, 3) - points.view(-1, 1, 3)
    pt_cross = torch.sum(pt_cross**2, dim=2)
    pt_cross = torch.exp(-pt_cross / (0.5 * 0.3))
    scalar = 1.0 / torch.sum(pt_cross, dim=1)
    scalar = num_points * scalar / torch.sum(scalar)
    return points * (scalar.view(-1, 1))


def CalcSpherePointsRandom(num_points: int, device) -> torch.Tensor:
    pts = 2.0 * (torch.rand(num_points, 3, device=device) - 0.5)
    radius = torch.sum(pts**2, dim=1)
    while torch.max(radius) > 1.0:
        replace_pts = 2.0 * (torch.rand(num_points, 3, device=device) - 0.5)
        replace_mask = radius.gt(0.99)
        pts.masked_scatter_(replace_mask.view(-1, 1).repeat(1, 3), replace_pts)
        radius = torch.sum(pts**2, dim=1)

    return pts / radius.view(-1, 1)



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.
"""

from __future__ import annotations

import os
import tempfile
import uuid
from dataclasses import dataclass, field
from typing import Optional

import clusterscope

from fairchem.core.common.utils import (
    StrEnum,
    get_commit_hash,
    get_timestamp_uid,
)

ALLOWED_TOP_LEVEL_KEYS = {"job", "runner", "reducer"}

LOG_DIR_NAME = "logs"
CHECKPOINT_DIR_NAME = "checkpoints"
RESULTS_DIR = "results"
CONFIG_FILE_NAME = "canonical_config.yaml"
PREEMPTION_STATE_DIR_NAME = "preemption_state"


class SchedulerType(StrEnum):
    LOCAL = "local"
    SLURM = "slurm"


class DeviceType(StrEnum):
    CPU = "cpu"
    CUDA = "cuda"


class RunType(StrEnum):
    RUN = "run"
    REDUCE = "reduce"


class DistributedInitMethod(StrEnum):
    TCP = "tcp"
    FILE = "file"


@dataclass
class SlurmConfig:
    mem_gb: int = 80
    timeout_hr: int = 168
    cpus_per_task: int = 8
    partition: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    qos: Optional[str] = None  # omegaconf in python 3.9 does not backport annotations
    account: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    additional_parameters: Optional[dict] = None


@dataclass
class RayClusterConfig:
    head_gpus: int = 0


@dataclass
class SchedulerConfig:
    mode: SchedulerType = SchedulerType.LOCAL
    distributed_init_method: DistributedInitMethod = DistributedInitMethod.TCP
    ranks_per_node: int = 1
    num_nodes: int = 1
    num_array_jobs: int = 1
    slurm: SlurmConfig = field(default_factory=lambda: SlurmConfig())
    # if not None, will launch a ray cluster on slurm instead of using submitit directly to launch the job
    use_ray: bool = False
    ray_cluster: RayClusterConfig = field(default_factory=lambda: RayClusterConfig())


@dataclass
class SlurmEnv:
    # reflects the job_id given by submitit (slurm id with array job id and array task id if they exist)
    job_id: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # reflects SLURM_JOB_ID only
    raw_job_id: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # SLURM_ARRAY_JOB_ID
    array_job_id: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # SLURM_ARRAY_TASK_ID
    array_task_id: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # reflects SLURM_RESTART_COUNT env variable
    restart_count: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )


@dataclass
class Metadata:
    # read-only metadata about the job, not user inputs
    commit: str
    log_dir: str
    checkpoint_dir: str
    results_dir: str
    config_path: str
    preemption_checkpoint_dir: str
    cluster_name: str
    array_job_num: int = 0
    slurm_env: SlurmEnv = field(default_factory=lambda: SlurmEnv())


@dataclass
class JobConfig:
    run_name: str = field(
        default_factory=lambda: get_timestamp_uid() + uuid.uuid4().hex.upper()[0:4]
    )
    timestamp_id: str = field(default_factory=lambda: get_timestamp_uid())
    run_dir: str = field(default_factory=lambda: tempfile.TemporaryDirectory().name)
    device_type: DeviceType = DeviceType.CUDA
    debug: bool = False
    scheduler: SchedulerConfig = field(default_factory=lambda: SchedulerConfig)
    logger: Optional[dict] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    seed: int = 0
    deterministic: bool = False
    runner_state_path: Optional[str] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    # read-only metadata about the job, not user inputs
    metadata: Optional[Metadata] = (
        None  # omegaconf in python 3.9 does not backport annotations
    )
    graph_parallel_group_size: Optional[int] = None

    def __post_init__(self) -> None:
        self.run_dir = os.path.abspath(self.run_dir)
        try:
            cluster = clusterscope.cluster()
        except RuntimeError:
            cluster = ""
        self.metadata = Metadata(
            commit=get_commit_hash(),
            log_dir=os.path.join(self.run_dir, self.timestamp_id, LOG_DIR_NAME),
            checkpoint_dir=os.path.join(
                self.run_dir, self.timestamp_id, CHECKPOINT_DIR_NAME
            ),
            results_dir=os.path.join(self.run_dir, self.timestamp_id, RESULTS_DIR),
            config_path=os.path.join(self.run_dir, self.timestamp_id, CONFIG_FILE_NAME),
            preemption_checkpoint_dir=os.path.join(
                self.run_dir,
                self.timestamp_id,
                CHECKPOINT_DIR_NAME,
                PREEMPTION_STATE_DIR_NAME,
            ),
            cluster_name=cluster,
        )



"""
Copyright (c) Meta Platforms, Inc. and affiliates.

This source code is licensed under the MIT license found in the
LICENSE file in the root directory of this source tree.

Add `extra_repr` into DropPath implemented by timm
for displaying more info.
"""

from __future__ import annotations

import torch
import torch.nn as nn
import torch.nn.functional as F
from e3nn import o3


def drop_path(
    x: torch.Tensor, drop_prob: float = 0.0, training: bool = False
) -> torch.Tensor:
    """Drop paths (Stochastic Depth) per sample (when applied in main path of residual blocks).
    This is the same as the DropConnect impl I created for EfficientNet, etc networks, however,
    the original name is misleading as 'Drop Connect' is a different form of dropout in a separate paper...
    See discussion: https://github.com/tensorflow/tpu/issues/494#issuecomment-532968956 ... I've opted for
    changing the layer and argument names to 'drop path' rather than mix DropConnect as a layer name and use
    'survival rate' as the argument.
    """
    if drop_prob == 0.0 or not training:
        return x
    keep_prob = 1 - drop_prob
    shape = (x.shape[0],) + (1,) * (
        x.ndim - 1
    )  # work with diff dim tensors, not just 2D ConvNets
    random_tensor = keep_prob + torch.rand(shape, dtype=x.dtype, device=x.device)
    random_tensor.floor_()  # binarize
    return x.div(keep_prob) * random_tensor


class DropPath(nn.Module):
    """Drop paths (Stochastic Depth) per sample  (when applied in main path of residual blocks)."""

    def __init__(self, drop_prob: float) -> None:
        super().__init__()
        self.drop_prob = drop_prob

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return drop_path(x, self.drop_prob, self.training)

    def extra_repr(self) -> str:
        return f"drop_prob={self.drop_prob}"


class GraphDropPath(nn.Module):
    """
    Consider batch for graph data when dropping paths.
    """

    def __init__(self, drop_prob: float) -> None:
        super().__init__()
        self.drop_prob = drop_prob

    def forward(self, x: torch.Tensor, batch) -> torch.Tensor:
        batch_size = batch.max() + 1
        shape = (batch_size,) + (1,) * (
            x.ndim - 1
        )  # work with diff dim tensors, not just 2D ConvNets
        ones = torch.ones(shape, dtype=x.dtype, device=x.device)
        drop = drop_path(ones, self.drop_prob, self.training)
        return x * drop[batch]

    def extra_repr(self) -> str:
        return f"drop_prob={self.drop_prob}"


class EquivariantDropout(nn.Module):
    def __init__(self, irreps, drop_prob: float) -> None:
        super().__init__()
        self.irreps = irreps
        self.num_irreps = irreps.num_irreps
        self.drop_prob = drop_prob
        self.drop = torch.nn.Dropout(drop_prob, True)
        self.mul = o3.ElementwiseTensorProduct(
            irreps, o3.Irreps(f"{self.num_irreps}x0e")
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training or self.drop_prob == 0.0:
            return x
        shape = (x.shape[0], self.num_irreps)
        mask = torch.ones(shape, dtype=x.dtype, device=x.device)
        mask = self.drop(mask)
        return self.mul(x, mask)


class EquivariantScalarsDropout(nn.Module):
    def __init__(self, irreps, drop_prob: float) -> None:
        super().__init__()
        self.irreps = irreps
        self.drop_prob = drop_prob

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self.training or self.drop_prob == 0.0:
            return x
        out = []
        start_idx = 0
        for mul, ir in self.irreps:
            temp = x.narrow(-1, start_idx, mul * ir.dim)
            start_idx += mul * ir.dim
            if ir.is_scalar():
                temp = F.dropout(temp, p=self.drop_prob, training=self.training)
            out.append(temp)
        return torch.cat(out, dim=-1)

    def extra_repr(self) -> str:
        return f"irreps={self.irreps}, drop_prob={self.drop_prob}"


class EquivariantDropoutArraySphericalHarmonics(nn.Module):
    def __init__(self, drop_prob: float, drop_graph: bool = False) -> None:
        super().__init__()
        self.drop_prob = drop_prob
        self.drop = torch.nn.Dropout(drop_prob, True)
        self.drop_graph = drop_graph

    def forward(self, x: torch.Tensor, batch=None) -> torch.Tensor:
        if not self.training or self.drop_prob == 0.0:
            return x
        assert len(x.shape) == 3

        if self.drop_graph:
            assert batch is not None
            batch_size = batch.max() + 1
            shape = (batch_size, 1, x.shape[2])
            mask = torch.ones(shape, dtype=x.dtype, device=x.device)
            mask = self.drop(mask)
            out = x * mask[batch]
        else:
            shape = (x.shape[0], 1, x.shape[2])
            mask = torch.ones(shape, dtype=x.dtype, device=x.device)
            mask = self.drop(mask)
            out = x * mask

        return out

    def extra_repr(self) -> str:
        return f"drop_prob={self.drop_prob}, drop_graph={self.drop_graph}"
